package gr.cite.annotation.web.scope.user;


import gr.cite.annotation.common.scope.user.UserScope;
import gr.cite.annotation.data.UserCredentialEntity;
import gr.cite.annotation.model.UserCredential;
import gr.cite.annotation.query.UserCredentialQuery;
import gr.cite.commons.web.oidc.principal.CurrentPrincipalResolver;
import gr.cite.commons.web.oidc.principal.extractor.ClaimExtractor;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyForbiddenException;
import gr.cite.tools.fieldset.BaseFieldSet;
import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Component;
import org.springframework.ui.ModelMap;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.context.request.WebRequestInterceptor;

import java.util.UUID;

@Component
public class UserInterceptor implements WebRequestInterceptor {
    private final UserScope userScope;
    private final ClaimExtractor claimExtractor;
    private final CurrentPrincipalResolver currentPrincipalResolver;
    private final UserInterceptorCacheService userInterceptorCacheService;
    private final QueryFactory queryFactory;

    @Autowired
    public UserInterceptor(
		    UserScope userScope,
		    ClaimExtractor claimExtractor,
		    CurrentPrincipalResolver currentPrincipalResolver,
		    UserInterceptorCacheService userInterceptorCacheService,
		    QueryFactory queryFactory) {
        this.userScope = userScope;
        this.currentPrincipalResolver = currentPrincipalResolver;
        this.claimExtractor = claimExtractor;
        this.userInterceptorCacheService = userInterceptorCacheService;
        this.queryFactory = queryFactory;
    }

    @Override
    public void preHandle(@NotNull WebRequest request) {
        UUID userId = null;
        if (this.currentPrincipalResolver.currentPrincipal().isAuthenticated()) {
            String subjectId = this.claimExtractor.subjectString(this.currentPrincipalResolver.currentPrincipal());
            if (subjectId == null || subjectId.isBlank()) throw new MyForbiddenException("Empty subjects not allowed");
            
            UserInterceptorCacheService.UserInterceptorCacheValue cacheValue = this.userInterceptorCacheService.lookup(this.userInterceptorCacheService.buildKey(subjectId));
            if (cacheValue != null) {
                userId = cacheValue.getUserId();
            } else {
                userId = this.findExistingUserFromDb(subjectId);
                if (userId != null) {
                    cacheValue = new UserInterceptorCacheService.UserInterceptorCacheValue(subjectId, userId);
                    this.userInterceptorCacheService.put(cacheValue);
                }
            }
        }
        this.userScope.setUserId(userId);
    }
    private UUID findExistingUserFromDb(String subjectId) {
        UserCredentialEntity userCredential = this.queryFactory.query(UserCredentialQuery.class).externalIds(subjectId).firstAs(new BaseFieldSet().ensure(UserCredential._user));
        if (userCredential != null) {
            return userCredential.getUserId();
        }
        return null;
    }
    @Override
    public void postHandle(@NonNull WebRequest request, ModelMap model) {
        this.userScope.setUserId(null);
    }

    @Override
    public void afterCompletion(@NonNull WebRequest request, Exception ex) {
    }
}

