package gr.cite.annotation.integrationevent.inbox.annotationentitiestouch;

import gr.cite.annotation.audit.AuditableAction;
import gr.cite.annotation.common.JsonHandlingService;
import gr.cite.annotation.common.enums.IsActive;
import gr.cite.annotation.common.scope.fake.FakeRequestScope;
import gr.cite.annotation.common.scope.tenant.TenantScope;
import gr.cite.annotation.data.EntityUserEntity;
import gr.cite.annotation.data.TenantEntity;
import gr.cite.annotation.data.TenantEntityManager;
import gr.cite.annotation.integrationevent.inbox.EventProcessingStatus;
import gr.cite.annotation.integrationevent.inbox.InboxPrincipal;
import gr.cite.annotation.integrationevent.inbox.IntegrationEventProperties;
import gr.cite.annotation.model.Tenant;
import gr.cite.annotation.query.EntityUserQuery;
import gr.cite.annotation.query.TenantQuery;
import gr.cite.annotation.service.tenant.TenantService;
import gr.cite.commons.web.oidc.principal.CurrentPrincipalResolver;
import gr.cite.commons.web.oidc.principal.extractor.ClaimExtractorProperties;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.data.deleter.DeleterFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.validation.ValidatorFactory;
import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.EntityTransaction;
import jakarta.persistence.OptimisticLockException;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.management.InvalidApplicationException;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class AnnotationEntitiesTouchedIntegrationEventHandlerImpl implements AnnotationEntitiesTouchedIntegrationEventHandler {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(AnnotationEntitiesTouchedIntegrationEventHandlerImpl.class));

    private final JsonHandlingService jsonHandlingService;

    private final ValidatorFactory validatorFactory;

    private final QueryFactory queryFactory;
    private final CurrentPrincipalResolver currentPrincipalResolver;
    private final ClaimExtractorProperties claimExtractorProperties;
    private final AuditService auditService;
    private final TenantEntityManager tenantEntityManager;
    private final DeleterFactory deleterFactory;
    private final TenantScope tenantScope;

    public AnnotationEntitiesTouchedIntegrationEventHandlerImpl(JsonHandlingService jsonHandlingService, ValidatorFactory validatorFactory, QueryFactory queryFactory, CurrentPrincipalResolver currentPrincipalResolver, ClaimExtractorProperties claimExtractorProperties, AuditService auditService, TenantEntityManager tenantEntityManager, DeleterFactory deleterFactory, TenantScope tenantScope) {
        this.jsonHandlingService = jsonHandlingService;
        this.validatorFactory = validatorFactory;
        this.currentPrincipalResolver = currentPrincipalResolver;
        this.claimExtractorProperties = claimExtractorProperties;
        this.auditService = auditService;
        this.tenantEntityManager = tenantEntityManager;
        this.queryFactory = queryFactory;
	    this.deleterFactory = deleterFactory;
	    this.tenantScope = tenantScope;
    }

    @Override
    public EventProcessingStatus handle(IntegrationEventProperties properties, String message) {
        AnnotationEntitiesTouchedIntegrationEvent event = this.jsonHandlingService.fromJsonSafe(AnnotationEntitiesTouchedIntegrationEvent.class, message);
        if (event == null)
            return EventProcessingStatus.Error;

        logger.debug("Handling {}", AnnotationEntitiesTouchedIntegrationEvent.class.getSimpleName());

        this.validatorFactory.validator(AnnotationEntitiesTouchedIntegrationEvent.AnnotationEntitiesTouchedIntegrationEventValidator.class).validateForce(event);

        EventProcessingStatus status = EventProcessingStatus.Success;
        try {
            if (this.tenantScope.isMultitenant() && properties.getTenantId() != null) {
                TenantEntity tenant = queryFactory.query(TenantQuery.class).ids(properties.getTenantId()).firstAs(new BaseFieldSet().ensure(Tenant._id).ensure(Tenant._code));
                if (tenant == null) {
                    logger.error("missing tenant from event message");
                    return EventProcessingStatus.Error;
                }
                this.tenantScope.setTempTenant(tenantEntityManager.getEntityManager(), properties.getTenantId(), tenant.getCode());
            } else if (this.tenantScope.isMultitenant()) {
//                    logger.error("missing tenant from event message");
//                    return EventProcessingStatus.Error;
                this.tenantScope.setTempTenant(tenantEntityManager.getEntityManager(), null, this.tenantScope.getDefaultTenantCode());
            }

            currentPrincipalResolver.push(InboxPrincipal.build(properties, claimExtractorProperties));

            tenantEntityManager.disableTenantFilters();
            for (AnnotationEntitiesTouchedIntegrationEvent.AnnotationEntityTouchedIntegrationEvent entityEvent : event.getEvents()) {

                EntityUserQuery entityUserQuery = this.queryFactory.query(EntityUserQuery.class);
                List<EntityUserEntity> items = entityUserQuery
                        .entityIds(entityEvent.getEntityId())
                        .isActive(IsActive.Active)
                        .collect();
                List<UUID> updatedCreatedIds = new ArrayList<>();
                for (UUID user : entityEvent.getUserIds()) {
                    EntityUserEntity data = items.stream().filter(x -> x.getUserId().equals(user)).findFirst().orElse(null);
                    if (data == null) {
                        data = new EntityUserEntity();
                        data.setId(UUID.randomUUID());
                        data.setEntityId(entityEvent.getEntityId());
                        data.setUserId(user);
                        data.setTenantId(properties.getTenantId());
                        data.setCreatedAt(Instant.now());
                        data.setUpdatedAt(Instant.now());
                        data.setIsActive(IsActive.Active);
                        tenantEntityManager.persist(data);
                    }
                    updatedCreatedIds.add(data.getId());
                }

                List<EntityUserEntity> toDelete = items.stream().filter(x -> updatedCreatedIds.stream().noneMatch(y -> y.equals(x.getId()))).collect(Collectors.toList());
                deleterFactory.deleter(gr.cite.EntityUser.model.deleter.EntityUserDeleter.class).delete(toDelete);

                tenantEntityManager.flush();


                auditService.track(AuditableAction.User_Persist, Map.ofEntries(
                        new AbstractMap.SimpleEntry<String, Object>("model", event)
                ));
            }
            
        } catch (Exception ex) {
            status = EventProcessingStatus.Error;
            logger.error("Problem getting list of queue outbox. Skipping: {}", ex.getMessage(), ex);
        } finally {
            currentPrincipalResolver.pop();
            tenantScope.removeTempTenant(this.tenantEntityManager.getEntityManager());
            try {
                this.tenantEntityManager.enableTenantFilters();
            } catch (InvalidApplicationException e) {
            }
        }

        return status;
    }

}
