package eu.eudat.model.builder;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.convention.ConventionService;
import eu.eudat.data.DmpReferenceEntity;
import eu.eudat.model.PublicDmp;
import eu.eudat.model.PublicDmpReference;
import eu.eudat.model.PublicReference;
import eu.eudat.query.DmpQuery;
import eu.eudat.query.ReferenceQuery;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.DataLogEntry;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class PublicDmpReferenceBuilder extends BaseBuilder<PublicDmpReference, DmpReferenceEntity> {

    private final BuilderFactory builderFactory;

    private final QueryFactory queryFactory;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public PublicDmpReferenceBuilder(
            ConventionService conventionService,
            BuilderFactory builderFactory, QueryFactory queryFactory) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(PublicDmpReferenceBuilder.class)));
        this.builderFactory = builderFactory;
        this.queryFactory = queryFactory;
    }

    public PublicDmpReferenceBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    public List<PublicDmpReference> build(FieldSet fields, List<DmpReferenceEntity> data) throws MyApplicationException {
        this.logger.debug("building for {} items requesting {} fields", Optional.ofNullable(data).map(List::size).orElse(0), Optional.ofNullable(fields).map(FieldSet::getFields).map(Set::size).orElse(0));
        this.logger.trace(new DataLogEntry("requested fields", fields));
        if (fields == null || data == null || fields.isEmpty())
            return new ArrayList<>();

        FieldSet referenceFields = fields.extractPrefixed(this.asPrefix(PublicDmpReference._reference));
        Map<UUID, PublicReference> referenceItemsMap = this.collectReferences(referenceFields, data);

        FieldSet dmpFields = fields.extractPrefixed(this.asPrefix(PublicDmpReference._dmp));
        Map<UUID, PublicDmp> dmpItemsMap = this.collectDmps(dmpFields, data);

        List<PublicDmpReference> models = new ArrayList<>();
        for (DmpReferenceEntity d : data) {
            PublicDmpReference m = new PublicDmpReference();
            if (fields.hasField(this.asIndexer(PublicDmpReference._id))) m.setId(d.getId());
            if (fields.hasField(this.asIndexer(PublicDmpReference._isActive))) m.setIsActive(d.getIsActive());
            if (!referenceFields.isEmpty() && referenceItemsMap != null && referenceItemsMap.containsKey(d.getReferenceId()))  m.setReference(referenceItemsMap.get(d.getReferenceId()));
            if (!dmpFields.isEmpty() && dmpItemsMap != null && dmpItemsMap.containsKey(d.getDmpId())) m.setDmp(dmpItemsMap.get(d.getDmpId()));
            models.add(m);
        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));
        return models;
    }

    private Map<UUID, PublicReference> collectReferences(FieldSet fields, List<DmpReferenceEntity> data) throws MyApplicationException {
        if (fields.isEmpty() || data.isEmpty())
            return null;
        this.logger.debug("checking related - {}", PublicReference.class.getSimpleName());

        Map<UUID, PublicReference> itemMap;
        if (!fields.hasOtherField(this.asIndexer(PublicReference._id))) {
            itemMap = this.asEmpty(
                    data.stream().map(DmpReferenceEntity::getReferenceId).distinct().collect(Collectors.toList()),
                    x -> {
                        PublicReference item = new PublicReference();
                        item.setId(x);
                        return item;
                    },
                    PublicReference::getId);
        } else {
            FieldSet clone = new BaseFieldSet(fields.getFields()).ensure(PublicReference._id);
            ReferenceQuery q = this.queryFactory.query(ReferenceQuery.class).authorize(this.authorize).isActive(IsActive.Active).ids(data.stream().map(DmpReferenceEntity::getReferenceId).distinct().collect(Collectors.toList()));
            itemMap = this.builderFactory.builder(PublicReferenceBuilder.class).authorize(this.authorize).asForeignKey(q, clone, PublicReference::getId);
        }
        if (!fields.hasField(PublicReference._id)) {
            itemMap.values().stream().filter(Objects::nonNull).peek(x -> x.setId(null)).collect(Collectors.toList());
        }

        return itemMap;
    }

    private Map<UUID, PublicDmp> collectDmps(FieldSet fields, List<DmpReferenceEntity> data) throws MyApplicationException {
        if (fields.isEmpty() || data.isEmpty())
            return null;
        this.logger.debug("checking related - {}", PublicDmp.class.getSimpleName());

        Map<UUID, PublicDmp> itemMap;
        if (!fields.hasOtherField(this.asIndexer(PublicDmp._id))) {
            itemMap = this.asEmpty(
                    data.stream().map(DmpReferenceEntity::getDmpId).distinct().collect(Collectors.toList()),
                    x -> {
                        PublicDmp item = new PublicDmp();
                        item.setId(x);
                        return item;
                    },
                    PublicDmp::getId);
        } else {
            FieldSet clone = new BaseFieldSet(fields.getFields()).ensure(PublicDmp._id);
            DmpQuery q = this.queryFactory.query(DmpQuery.class).authorize(this.authorize).isActive(IsActive.Active).ids(data.stream().map(DmpReferenceEntity::getDmpId).distinct().collect(Collectors.toList()));
            itemMap = this.builderFactory.builder(PublicDmpBuilder.class).authorize(this.authorize).asForeignKey(q, clone, PublicDmp::getId);
        }
        if (!fields.hasField(PublicDmp._id)) {
            itemMap.values().stream().filter(Objects::nonNull).peek(x -> x.setId(null)).collect(Collectors.toList());
        }

        return itemMap;
    }

}
