package eu.eudat.model.builder;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.convention.ConventionService;
import eu.eudat.data.DmpUserEntity;
import eu.eudat.model.PublicDmp;
import eu.eudat.model.PublicDmpUser;
import eu.eudat.model.PublicUser;
import eu.eudat.model.User;
import eu.eudat.query.DmpQuery;
import eu.eudat.query.UserQuery;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.DataLogEntry;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class PublicDmpUserBuilder extends BaseBuilder<PublicDmpUser, DmpUserEntity>{

    private final BuilderFactory builderFactory;

    private final QueryFactory queryFactory;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public PublicDmpUserBuilder(
            ConventionService conventionService,
            BuilderFactory builderFactory, QueryFactory queryFactory) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(PublicDmpUserBuilder.class)));
        this.builderFactory = builderFactory;
        this.queryFactory = queryFactory;
    }

    public PublicDmpUserBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    public List<PublicDmpUser> build(FieldSet fields, List<DmpUserEntity> data) throws MyApplicationException {
        this.logger.debug("building for {} items requesting {} fields", Optional.ofNullable(data).map(List::size).orElse(0), Optional.ofNullable(fields).map(FieldSet::getFields).map(Set::size).orElse(0));
        this.logger.trace(new DataLogEntry("requested fields", fields));
        if (fields == null || data == null || fields.isEmpty())
            return new ArrayList<>();

        FieldSet userFields = fields.extractPrefixed(this.asPrefix(PublicDmpUser._user));
        Map<UUID, PublicUser> userItemsMap = this.collectUsers(userFields, data);

        FieldSet dmpFields = fields.extractPrefixed(this.asPrefix(PublicDmpUser._dmp));
        Map<UUID, PublicDmp> dmpItemsMap = this.collectDmps(dmpFields, data);

        List<PublicDmpUser> models = new ArrayList<>();
        for (DmpUserEntity d : data) {
            PublicDmpUser m = new PublicDmpUser();
            if (fields.hasField(this.asIndexer(PublicDmpUser._id))) m.setId(d.getId());
            if (fields.hasField(this.asIndexer(PublicDmpUser._role))) m.setRole(d.getRole());
            if (!userFields.isEmpty() && userItemsMap != null && userItemsMap.containsKey(d.getUserId())) m.setUser(userItemsMap.get(d.getUserId()));
            if (!dmpFields.isEmpty() && dmpItemsMap != null && dmpItemsMap.containsKey(d.getDmpId())) m.setDmp(dmpItemsMap.get(d.getDmpId()));
            models.add(m);
        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));
        return models;
    }

    private Map<UUID, PublicUser> collectUsers(FieldSet fields, List<DmpUserEntity> data) throws MyApplicationException {
        if (fields.isEmpty() || data.isEmpty())
            return null;
        this.logger.debug("checking related - {}", User.class.getSimpleName());

        Map<UUID, PublicUser> itemMap;
        if (!fields.hasOtherField(this.asIndexer(PublicUser._id))) {
            itemMap = this.asEmpty(
                    data.stream().map(DmpUserEntity::getUserId).distinct().collect(Collectors.toList()),
                    x -> {
                        PublicUser item = new PublicUser();
                        item.setId(x);
                        return item;
                    },
                    PublicUser::getId);
        } else {
            FieldSet clone = new BaseFieldSet(fields.getFields()).ensure(PublicUser._id);
            UserQuery q = this.queryFactory.query(UserQuery.class).authorize(this.authorize).ids(data.stream().map(DmpUserEntity::getUserId).distinct().collect(Collectors.toList()));
            itemMap = this.builderFactory.builder(PublicUserBuilder.class).authorize(this.authorize).asForeignKey(q, clone, PublicUser::getId);
        }
        if (!fields.hasField(PublicUser._id)) {
            itemMap.forEach((id, item) -> {
                if (item != null)
                    item.setId(null);
            });
        }

        return itemMap;
    }

    private Map<UUID, PublicDmp> collectDmps(FieldSet fields, List<DmpUserEntity> data) throws MyApplicationException {
        if (fields.isEmpty() || data.isEmpty())
            return null;
        this.logger.debug("checking related - {}", PublicDmp.class.getSimpleName());

        Map<UUID, PublicDmp> itemMap;
        if (!fields.hasOtherField(this.asIndexer(PublicDmp._id))) {
            itemMap = this.asEmpty(
                    data.stream().map(DmpUserEntity::getDmpId).distinct().collect(Collectors.toList()),
                    x -> {
                        PublicDmp item = new PublicDmp();
                        item.setId(x);
                        return item;
                    },
                    PublicDmp::getId);
        } else {
            FieldSet clone = new BaseFieldSet(fields.getFields()).ensure(PublicDmp._id);
            DmpQuery q = this.queryFactory.query(DmpQuery.class).authorize(this.authorize).ids(data.stream().map(DmpUserEntity::getDmpId).distinct().collect(Collectors.toList()));
            itemMap = this.builderFactory.builder(PublicDmpBuilder.class).authorize(this.authorize).asForeignKey(q, clone, PublicDmp::getId);
        }
        if (!fields.hasField(PublicDmp._id)) {
            itemMap.values().stream().filter(Objects::nonNull).peek(x -> x.setId(null)).collect(Collectors.toList());
        }

        return itemMap;
    }

}
