package eu.eudat.model.deleter;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.enums.DmpStatus;
import eu.eudat.commons.enums.DmpVersionStatus;
import eu.eudat.commons.enums.EntityType;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.data.*;
import eu.eudat.model.Description;
import eu.eudat.model.DmpDescriptionTemplate;
import eu.eudat.model.DmpReference;
import eu.eudat.query.*;
import eu.eudat.service.elastic.ElasticService;
import gr.cite.tools.data.deleter.Deleter;
import gr.cite.tools.data.deleter.DeleterFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.logging.MapLogEntry;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.management.InvalidApplicationException;
import java.io.IOException;
import java.time.Instant;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class DmpDeleter implements Deleter {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(DmpDeleter.class));

    private final TenantEntityManager entityManager;

    protected final QueryFactory queryFactory;

    protected final DeleterFactory deleterFactory;
    protected final ElasticService elasticService;

    @Autowired
    public DmpDeleter(
            TenantEntityManager entityManager,
            QueryFactory queryFactory,
            DeleterFactory deleterFactory,
            ElasticService elasticService) {
        this.entityManager = entityManager;
        this.queryFactory = queryFactory;
        this.deleterFactory = deleterFactory;
        this.elasticService = elasticService;
    }

    public void deleteAndSaveByIds(List<UUID> ids, boolean disableElastic) throws InvalidApplicationException, IOException {
        logger.debug(new MapLogEntry("collecting to delete").And("count", Optional.ofNullable(ids).map(List::size).orElse(0)).And("ids", ids));
        List<DmpEntity> data = this.queryFactory.query(DmpQuery.class).ids(ids).collect();
        logger.trace("retrieved {} items", Optional.ofNullable(data).map(List::size).orElse(0));
        this.deleteAndSave(data, disableElastic);
    }

    public void deleteAndSave(List<DmpEntity> data, boolean disableElastic) throws InvalidApplicationException, IOException {
        logger.debug("will delete {} items", Optional.ofNullable(data).map(List::size).orElse(0));
        this.delete(data, disableElastic);
        logger.trace("saving changes");
        this.entityManager.flush();
        logger.trace("changes saved");
    }

    public void delete(List<DmpEntity> data, boolean disableElastic) throws InvalidApplicationException, IOException {
        logger.debug("will delete {} items", Optional.ofNullable(data).map(List::size).orElse(0));
        if (data == null || data.isEmpty())
            return;

        List<UUID> ids = data.stream().map(DmpEntity::getId).distinct().toList();
        {
            logger.debug("checking related - {}", DmpUserEntity.class.getSimpleName());
            List<DmpUserEntity> items = this.queryFactory.query(DmpUserQuery.class).dmpIds(ids).collect();
            DmpUserDeleter deleter = this.deleterFactory.deleter(DmpUserDeleter.class);
            deleter.delete(items);
        }
        {
            logger.debug("checking related - {}", DmpDescriptionTemplate.class.getSimpleName());
            List<DmpDescriptionTemplateEntity> items = this.queryFactory.query(DmpDescriptionTemplateQuery.class).dmpIds(ids).collect();
            DmpDescriptionTemplateDeleter deleter = this.deleterFactory.deleter(DmpDescriptionTemplateDeleter.class);
            deleter.delete(items);
        }
        {
            logger.debug("checking related - {}", DmpReference.class.getSimpleName());
            List<DmpReferenceEntity> items = this.queryFactory.query(DmpReferenceQuery.class).dmpIds(ids).collect();
            DmpReferenceDeleter deleter = this.deleterFactory.deleter(DmpReferenceDeleter.class);
            deleter.delete(items);
        }
        {
            logger.debug("checking related - {}", Description.class.getSimpleName());
            List<DescriptionEntity> items = this.queryFactory.query(DescriptionQuery.class).dmpIds(ids).collect();
            DescriptionDeleter deleter = this.deleterFactory.deleter(DescriptionDeleter.class);
            deleter.delete(items, false); //We delete elastic entities by bmp deleter 
        }

        Instant now = Instant.now();

        for (DmpEntity item : data) {
            logger.trace("deleting item {}", item.getId());
            EntityDoiQuery entityDoiQuery = this.queryFactory.query(EntityDoiQuery.class).authorize(AuthorizationFlags.OwnerOrDmpAssociatedOrPermission).types(EntityType.DMP).entityIds(item.getId());
            if (entityDoiQuery.count() > 0) throw new MyApplicationException("DMP is deposited can not deleted");
            if(item.getVersionStatus().equals(DmpVersionStatus.Current)) throw new MyApplicationException("DMP is current can not deleted");
            item.setIsActive(IsActive.Inactive);
            item.setUpdatedAt(now);
            logger.trace("updating item");
            this.entityManager.merge(item);
            logger.trace("updated item");

            if (!disableElastic) this.elasticService.deleteDmp(item);
        }
    }

}
