package eu.eudat.query;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.authorization.Permission;
import eu.eudat.commons.enums.DescriptionStatus;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.commons.scope.user.UserScope;
import eu.eudat.data.DescriptionEntity;
import eu.eudat.data.DmpDescriptionTemplateEntity;
import eu.eudat.data.DmpEntity;
import eu.eudat.model.Description;
import eu.eudat.model.PublicDescription;
import eu.eudat.query.utils.QueryUtilsService;
import gr.cite.commons.web.authz.service.AuthorizationService;
import gr.cite.tools.data.query.FieldResolver;
import gr.cite.tools.data.query.QueryBase;
import gr.cite.tools.data.query.QueryContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.Predicate;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class DescriptionQuery extends QueryBase<DescriptionEntity> {

    private String like;

    private Collection<UUID> ids;

    private Collection<UUID> createdByIds;

    private DmpDescriptionTemplateQuery dmpDescriptionTemplateQuery;

    private DmpQuery dmpQuery;

    private Instant createdAfter;

    private Instant createdBefore;

    private Instant finalizedAfter;

    private Instant finalizedBefore;

    private Collection<UUID> excludedIds;

    private Collection<IsActive> isActives;

    private Collection<DescriptionStatus> statuses;

    private Collection<UUID> dmpIds;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    private final UserScope userScope;

    private final AuthorizationService authService;

    private final QueryUtilsService queryUtilsService;

    private Collection<UUID> dmpDescriptionTemplateIds;

    public DescriptionQuery(UserScope userScope, AuthorizationService authService, QueryUtilsService queryUtilsService) {
        this.userScope = userScope;
        this.authService = authService;
        this.queryUtilsService = queryUtilsService;
    }

    public DescriptionQuery like(String value) {
        this.like = value;
        return this;
    }

    public DescriptionQuery ids(UUID value) {
        this.ids = List.of(value);
        return this;
    }

    public DescriptionQuery ids(UUID... value) {
        this.ids = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery ids(Collection<UUID> values) {
        this.ids = values;
        return this;
    }

    public DescriptionQuery createdByIds(UUID value) {
        this.createdByIds = List.of(value);
        return this;
    }

    public DescriptionQuery createdByIds(UUID... value) {
        this.createdByIds = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery createdByIds(Collection<UUID> values) {
        this.createdByIds = values;
        return this;
    }

    public DescriptionQuery dmpDescriptionTemplateSubQuery(DmpDescriptionTemplateQuery subQuery) {
        this.dmpDescriptionTemplateQuery = subQuery;
        return this;
    }

    public DescriptionQuery createdAfter(Instant value) {
        this.createdAfter = value;
        return this;
    }

    public DescriptionQuery createdBefore(Instant value) {
        this.createdBefore = value;
        return this;
    }

    public DescriptionQuery dmpDescriptionTemplateIds(UUID value) {
        this.dmpDescriptionTemplateIds = List.of(value);
        return this;
    }

    public DescriptionQuery dmpDescriptionTemplateIds(UUID... value) {
        this.dmpDescriptionTemplateIds = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery dmpDescriptionTemplateIds(Collection<UUID> values) {
        this.dmpDescriptionTemplateIds = values;
        return this;
    }

    public DescriptionQuery finalizedAfter(Instant value) {
        this.finalizedAfter = value;
        return this;
    }

    public DescriptionQuery finalizedBefore(Instant value) {
        this.finalizedBefore = value;
        return this;
    }

    public DescriptionQuery dmpSubQuery(DmpQuery subQuery) {
        this.dmpQuery = subQuery;
        return this;
    }

    public DescriptionQuery excludedIds(Collection<UUID> values) {
        this.excludedIds = values;
        return this;
    }

    public DescriptionQuery excludedIds(UUID value) {
        this.excludedIds = List.of(value);
        return this;
    }

    public DescriptionQuery excludedIds(UUID... value) {
        this.excludedIds = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery isActive(IsActive value) {
        this.isActives = List.of(value);
        return this;
    }

    public DescriptionQuery isActive(IsActive... value) {
        this.isActives = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery isActive(Collection<IsActive> values) {
        this.isActives = values;
        return this;
    }

    public DescriptionQuery statuses(DescriptionStatus value) {
        this.statuses = List.of(value);
        return this;
    }

    public DescriptionQuery statuses(DescriptionStatus... value) {
        this.statuses = Arrays.asList(value);
        return this;
    }

    public DescriptionQuery statuses(Collection<DescriptionStatus> values) {
        this.statuses = values;
        return this;
    }

    public DescriptionQuery dmpIds(Collection<UUID> values) {
        this.dmpIds = values;
        return this;
    }

    public DescriptionQuery dmpIds(UUID value) {
        this.dmpIds = List.of(value);
        return this;
    }

    public DescriptionQuery authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    protected Boolean isFalseQuery() {
        return
                this.isEmpty(this.ids) ||
                        this.isEmpty(this.isActives) || this.isEmpty(this.createdByIds) ||
                        this.isEmpty(this.excludedIds) || this.isFalseQuery(this.dmpQuery) ||
                        this.isEmpty(this.statuses) || this.isFalseQuery(this.dmpDescriptionTemplateQuery);
    }

    @Override
    protected Class<DescriptionEntity> entityClass() {
        return DescriptionEntity.class;
    }

    @Override
    protected <X, Y> Predicate applyAuthZ(QueryContext<X, Y> queryContext) {
        if (this.authorize.contains(AuthorizationFlags.None))
            return null;
        if (this.authorize.contains(AuthorizationFlags.Permission) && this.authService.authorize(Permission.BrowseDescription))
            return null;
        UUID userId;
        boolean usePublic = this.authorize.contains(AuthorizationFlags.Public);
        if (this.authorize.contains(AuthorizationFlags.DmpAssociated))
            userId = this.userScope.getUserIdSafe();
        else
            userId = null;

        List<Predicate> predicates = new ArrayList<>();
        if (userId != null || usePublic) {
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._id)).value(queryUtilsService.buildDescriptionAuthZSubQuery(queryContext.Query, queryContext.CriteriaBuilder, userId, usePublic)));
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return queryContext.CriteriaBuilder.or(); //Creates a false query
        }
    }

    @Override
    protected <X, Y> Predicate applyFilters(QueryContext<X, Y> queryContext) {
        List<Predicate> predicates = new ArrayList<>();
        if (this.like != null && !this.like.isEmpty()) {
            predicates.add(queryContext.CriteriaBuilder.or(
                    queryUtilsService.ilike(queryContext.CriteriaBuilder, queryContext.Root.get(DescriptionEntity._description), this.like),
                    queryUtilsService.ilike(queryContext.CriteriaBuilder, queryContext.Root.get(DescriptionEntity._label), this.like)
            ));
        }
        if (this.ids != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._id));
            for (UUID item : this.ids)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.createdByIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._createdById));
            for (UUID item : this.createdByIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.createdAfter != null) {
            predicates.add(queryContext.CriteriaBuilder.greaterThan(queryContext.Root.get(DescriptionEntity._createdAt), this.createdAfter));
        }
        if (this.createdBefore != null) {
            predicates.add(queryContext.CriteriaBuilder.lessThan(queryContext.Root.get(DescriptionEntity._createdAt), this.createdBefore));
        }
        if (this.finalizedAfter != null) {
            predicates.add(queryContext.CriteriaBuilder.greaterThan(queryContext.Root.get(DescriptionEntity._finalizedAt), this.finalizedAfter));
        }
        if (this.finalizedBefore != null) {
            predicates.add(queryContext.CriteriaBuilder.lessThan(queryContext.Root.get(DescriptionEntity._finalizedAt), this.finalizedAfter));
        }
        if (this.excludedIds != null) {
            CriteriaBuilder.In<UUID> notInClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._id));
            for (UUID item : this.excludedIds)
                notInClause.value(item);
            predicates.add(notInClause.not());
        }
        if (this.dmpIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._dmpId));
            for (UUID item : this.dmpIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.isActives != null) {
            CriteriaBuilder.In<IsActive> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._isActive));
            for (IsActive item : this.isActives)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.statuses != null) {
            CriteriaBuilder.In<DescriptionStatus> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._status));
            for (DescriptionStatus item : this.statuses)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.dmpDescriptionTemplateQuery != null) {
            QueryContext<DmpDescriptionTemplateEntity, UUID> subQuery = this.applySubQuery(this.dmpDescriptionTemplateQuery, queryContext, UUID.class, dmpDescriptionTemplateEntityRoot -> dmpDescriptionTemplateEntityRoot.get(DmpDescriptionTemplateEntity._id));
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._dmpDescriptionTemplateId)).value(subQuery.Query));
        }
        if (this.dmpQuery != null) {
            QueryContext<DmpEntity, UUID> subQuery = this.applySubQuery(this.dmpQuery, queryContext, UUID.class, dmpEntityRoot -> dmpEntityRoot.get(DmpEntity._id));
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._dmpId)).value(subQuery.Query));
        }
        if (this.dmpDescriptionTemplateIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DescriptionEntity._dmpDescriptionTemplateId));
            for (UUID item : this.dmpDescriptionTemplateIds)
                inClause.value(item);
            predicates.add(inClause);
        }

        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return null;
        }
    }

    @Override
    protected String fieldNameOf(FieldResolver item) {
        if (item.match(Description._id) || item.match(PublicDescription._id))
            return DescriptionEntity._id;
        else if (item.match(Description._label) || item.match(PublicDescription._label))
            return DescriptionEntity._label;
        else if (item.prefix(Description._properties))
            return DescriptionEntity._properties;
        else if (item.match(Description._status) || item.match(PublicDescription._status))
            return DescriptionEntity._status;
        else if (item.match(Description._description) || item.match(PublicDescription._description))
            return DescriptionEntity._description;
        else if (item.match(Description._createdBy))
            return DescriptionEntity._createdById;
        else if (item.prefix(Description._createdBy))
            return DescriptionEntity._createdById;
        else if (item.match(Description._createdAt) || item.match(PublicDescription._createdAt))
            return DescriptionEntity._createdAt;
        else if (item.match(Description._updatedAt) || item.match(PublicDescription._updatedAt))
            return DescriptionEntity._updatedAt;
        else if (item.match(Description._isActive))
            return DescriptionEntity._isActive;
        else if (item.match(Description._hash))
            return DescriptionEntity._updatedAt;
        else if (item.match(Description._finalizedAt) || item.match(PublicDescription._finalizedAt))
            return DescriptionEntity._finalizedAt;
        else if (item.prefix(Description._dmpDescriptionTemplate) || item.prefix(PublicDescription._dmpDescriptionTemplate))
            return DescriptionEntity._dmpDescriptionTemplateId;
        else if (item.match(Description._dmpDescriptionTemplate) || item.match(PublicDescription._dmpDescriptionTemplate))
            return DescriptionEntity._dmpDescriptionTemplateId;
        else if (item.prefix(Description._descriptionTemplate) || item.prefix(PublicDescription._descriptionTemplate))
            return DescriptionEntity._descriptionTemplateId;
        else if (item.match(Description._descriptionTemplate) || item.match(PublicDescription._descriptionTemplate))
            return DescriptionEntity._descriptionTemplateId;
        else if (item.prefix(Description._dmp))
            return DescriptionEntity._dmpId;
        else if (item.match(Description._dmp))
            return DescriptionEntity._dmpId;
        else if (item.match(Description._belongsToCurrentTenant))
            return DescriptionEntity._tenantId;
        else if (item.match(DescriptionEntity._tenantId))
            return DescriptionEntity._tenantId;
        else
            return null;
    }

    @Override
    protected DescriptionEntity convert(Tuple tuple, Set<String> columns) {
        DescriptionEntity item = new DescriptionEntity();
        item.setId(QueryBase.convertSafe(tuple, columns, DescriptionEntity._id, UUID.class));
        item.setTenantId(QueryBase.convertSafe(tuple, columns, DescriptionEntity._tenantId, UUID.class));
        item.setLabel(QueryBase.convertSafe(tuple, columns, DescriptionEntity._label, String.class));
        item.setProperties(QueryBase.convertSafe(tuple, columns, DescriptionEntity._properties, String.class));
        item.setStatus(QueryBase.convertSafe(tuple, columns, DescriptionEntity._status, DescriptionStatus.class));
        item.setDescription(QueryBase.convertSafe(tuple, columns, DescriptionEntity._description, String.class));
        item.setCreatedAt(QueryBase.convertSafe(tuple, columns, DescriptionEntity._createdAt, Instant.class));
        item.setUpdatedAt(QueryBase.convertSafe(tuple, columns, DescriptionEntity._updatedAt, Instant.class));
        item.setIsActive(QueryBase.convertSafe(tuple, columns, DescriptionEntity._isActive, IsActive.class));
        item.setFinalizedAt(QueryBase.convertSafe(tuple, columns, DescriptionEntity._finalizedAt, Instant.class));
        item.setDmpDescriptionTemplateId(QueryBase.convertSafe(tuple, columns, DescriptionEntity._dmpDescriptionTemplateId, UUID.class));
        item.setDescriptionTemplateId(QueryBase.convertSafe(tuple, columns, DescriptionEntity._descriptionTemplateId, UUID.class));
        item.setDmpId(QueryBase.convertSafe(tuple, columns, DescriptionEntity._dmpId, UUID.class));
        return item;
    }

}
