package eu.eudat.query;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.authorization.Permission;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.commons.scope.user.UserScope;
import eu.eudat.data.DmpEntity;
import eu.eudat.data.DmpReferenceEntity;
import eu.eudat.data.ReferenceEntity;
import eu.eudat.model.DmpReference;
import eu.eudat.model.PublicDmpReference;
import eu.eudat.query.utils.QueryUtilsService;
import gr.cite.commons.web.authz.service.AuthorizationService;
import gr.cite.tools.data.query.FieldResolver;
import gr.cite.tools.data.query.QueryBase;
import gr.cite.tools.data.query.QueryContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.Predicate;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class DmpReferenceQuery extends QueryBase<DmpReferenceEntity> {

    private Collection<UUID> ids;

    private Collection<IsActive> isActives;

    private Collection<UUID> dmpIds;

    private Collection<UUID> referenceIds;
    private DmpQuery dmpQuery;
    private ReferenceQuery referenceQuery;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);


    public DmpReferenceQuery ids(UUID value) {
        this.ids = List.of(value);
        return this;
    }

    public DmpReferenceQuery ids(UUID... value) {
        this.ids = Arrays.asList(value);
        return this;
    }

    public DmpReferenceQuery ids(Collection<UUID> values) {
        this.ids = values;
        return this;
    }

    public DmpReferenceQuery isActives(IsActive value) {
        this.isActives = List.of(value);
        return this;
    }

    public DmpReferenceQuery isActives(IsActive... value) {
        this.isActives = Arrays.asList(value);
        return this;
    }

    public DmpReferenceQuery isActives(Collection<IsActive> values) {
        this.isActives = values;
        return this;
    }

    public DmpReferenceQuery dmpIds(UUID value) {
        this.dmpIds = List.of(value);
        return this;
    }

    public DmpReferenceQuery dmpIds(UUID... value) {
        this.dmpIds = Arrays.asList(value);
        return this;
    }

    public DmpReferenceQuery dmpIds(Collection<UUID> values) {
        this.dmpIds = values;
        return this;
    }

    public DmpReferenceQuery referenceIds(UUID value) {
        this.referenceIds = List.of(value);
        return this;
    }

    public DmpReferenceQuery referenceIds(UUID... value) {
        this.referenceIds = Arrays.asList(value);
        return this;
    }

    public DmpReferenceQuery referenceIds(Collection<UUID> values) {
        this.referenceIds = values;
        return this;
    }

    public DmpReferenceQuery dmpSubQuery(DmpQuery value) {
        this.dmpQuery = value;
        return this;
    }

    public DmpReferenceQuery referenceSubQuery(ReferenceQuery value) {
        this.referenceQuery = value;
        return this;
    }

    public DmpReferenceQuery authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    private final UserScope userScope;

    private final AuthorizationService authService;
    private final QueryUtilsService queryUtilsService;

    public DmpReferenceQuery(
            UserScope userScope,
            AuthorizationService authService,
            QueryUtilsService queryUtilsService) {
        this.userScope = userScope;
        this.authService = authService;
        this.queryUtilsService = queryUtilsService;
    }

    @Override
    protected Class<DmpReferenceEntity> entityClass() {
        return DmpReferenceEntity.class;
    }

    @Override
    protected Boolean isFalseQuery() {
        return this.isEmpty(this.ids) || this.isEmpty(this.dmpIds) || this.isEmpty(this.referenceIds) || this.isFalseQuery(this.dmpQuery) || this.isFalseQuery(this.referenceQuery);
    }

    @Override
    protected <X, Y> Predicate applyAuthZ(QueryContext<X, Y> queryContext) {
        if (this.authorize.contains(AuthorizationFlags.None)) return null;
        if (this.authorize.contains(AuthorizationFlags.Permission) && this.authService.authorize(Permission.BrowseDmpReference)) return null;
        UUID userId;
        boolean usePublic = this.authorize.contains(AuthorizationFlags.Public);
        if (this.authorize.contains(AuthorizationFlags.DmpAssociated)) userId = this.userScope.getUserIdSafe();
        else  userId = null;

        List<Predicate> predicates = new ArrayList<>();
        if (userId != null || usePublic ) {
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._dmpId)).value(queryUtilsService.buildDmpAuthZSubQuery(queryContext.Query, queryContext.CriteriaBuilder, userId, usePublic)));
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return queryContext.CriteriaBuilder.or(); //Creates a false query
        }
    }
    
    @Override
    protected <X, Y> Predicate applyFilters(QueryContext<X, Y> queryContext) {
        List<Predicate> predicates = new ArrayList<>();
        if (this.ids != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._id));
            for (UUID item : this.ids)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.isActives != null) {
            CriteriaBuilder.In<IsActive> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._isActive));
            for (IsActive item : this.isActives)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.dmpIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._dmpId));
            for (UUID item : this.dmpIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.referenceIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._referenceId));
            for (UUID item : this.referenceIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.dmpQuery != null) {
            QueryContext<DmpEntity, UUID> subQuery = this.applySubQuery(this.dmpQuery, queryContext, UUID.class, dmpEntityRoot -> dmpEntityRoot.get(DmpEntity._id));
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._dmpId)).value(subQuery.Query));
        }
        if (this.referenceQuery != null) {
            QueryContext<ReferenceEntity, UUID> subQuery = this.applySubQuery(this.referenceQuery, queryContext, UUID.class, root -> root.get(ReferenceEntity._id));
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(DmpReferenceEntity._referenceId)).value(subQuery.Query));
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return null;
        }
    }

    @Override
    protected DmpReferenceEntity convert(Tuple tuple, Set<String> columns) {
        DmpReferenceEntity item = new DmpReferenceEntity();
        item.setId(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._id, UUID.class));
        item.setTenantId(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._tenantId, UUID.class));
        item.setDmpId(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._dmpId, UUID.class));
        item.setReferenceId(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._referenceId, UUID.class));
        item.setData(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._data, String.class));
        item.setIsActive(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._isActive, IsActive.class));
        item.setCreatedAt(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._createdAt, Instant.class));
        item.setUpdatedAt(QueryBase.convertSafe(tuple, columns, DmpReferenceEntity._updatedAt, Instant.class));
        return item;
    }

    @Override
    protected String fieldNameOf(FieldResolver item) {
        if (item.match(DmpReference._id) || item.match(PublicDmpReference._id)) return DmpReferenceEntity._id;
        else if (item.prefix(DmpReference._dmp) || item.prefix(PublicDmpReference._dmp)) return DmpReferenceEntity._dmpId;
        else if (item.prefix(DmpReference._reference) || item.prefix(PublicDmpReference._reference)) return DmpReferenceEntity._referenceId;
        else if (item.match(DmpReference._dmp) || item.match(PublicDmpReference._dmp)) return DmpReferenceEntity._dmpId;
        else if (item.match(DmpReference._reference) || item.match(PublicDmpReference._reference)) return DmpReferenceEntity._referenceId;
        else if (item.prefix(DmpReference._data)) return DmpReferenceEntity._data;
        else if (item.match(DmpReference._isActive)) return DmpReferenceEntity._isActive;
        else if (item.match(DmpReference._createdAt)) return DmpReferenceEntity._createdAt;
        else if (item.match(DmpReference._updatedAt)) return DmpReferenceEntity._updatedAt;
        else if (item.match(DmpReference._hash)) return DmpReferenceEntity._updatedAt;
        else if (item.match(DmpReference._belongsToCurrentTenant)) return DmpReferenceEntity._tenantId;
        else return null;
    }

}
