package eu.eudat.query;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.data.LanguageEntity;
import eu.eudat.model.Language;
import eu.eudat.query.utils.QueryUtilsService;
import gr.cite.tools.data.query.FieldResolver;
import gr.cite.tools.data.query.QueryBase;
import gr.cite.tools.data.query.QueryContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.Predicate;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class LanguageQuery extends QueryBase<LanguageEntity> {

    private String like;

    private Collection<UUID> ids;

    private Collection<IsActive> isActives;

    private Collection<String> codes;

    private Collection<UUID> excludedIds;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    public LanguageQuery like(String value) {
        this.like = value;
        return this;
    }

    public LanguageQuery ids(UUID value) {
        this.ids = List.of(value);
        return this;
    }

    public LanguageQuery ids(UUID... value) {
        this.ids = Arrays.asList(value);
        return this;
    }

    public LanguageQuery ids(Collection<UUID> values) {
        this.ids = values;
        return this;
    }

    public LanguageQuery isActive(IsActive value) {
        this.isActives = List.of(value);
        return this;
    }

    public LanguageQuery isActive(IsActive... value) {
        this.isActives = Arrays.asList(value);
        return this;
    }

    public LanguageQuery isActive(Collection<IsActive> values) {
        this.isActives = values;
        return this;
    }

    public LanguageQuery codes(String value) {
        this.codes = List.of(value);
        return this;
    }

    public LanguageQuery codes(String... value) {
        this.codes = Arrays.asList(value);
        return this;
    }

    public LanguageQuery codes(Collection<String> values) {
        this.codes = values;
        return this;
    }

    public LanguageQuery excludedIds(Collection<UUID> values) {
        this.excludedIds = values;
        return this;
    }

    public LanguageQuery excludedIds(UUID value) {
        this.excludedIds = List.of(value);
        return this;
    }

    public LanguageQuery excludedIds(UUID... value) {
        this.excludedIds = Arrays.asList(value);
        return this;
    }

    public LanguageQuery authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    private final QueryUtilsService queryUtilsService;
    public LanguageQuery(
		    QueryUtilsService queryUtilsService) {
	    this.queryUtilsService = queryUtilsService;
    }

    @Override
    protected Class<LanguageEntity> entityClass() {
        return LanguageEntity.class;
    }

    @Override
    protected Boolean isFalseQuery() {
        return this.isEmpty(this.ids) || this.isEmpty(this.isActives) || this.isEmpty(this.excludedIds) || this.isEmpty(this.codes);
    }

    @Override
    protected <X, Y> Predicate applyFilters(QueryContext<X, Y> queryContext) {
        List<Predicate> predicates = new ArrayList<>();
        if (this.ids != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(LanguageEntity._id));
            for (UUID item : this.ids)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.like != null && !this.like.isEmpty()) {
            predicates.add(queryUtilsService.ilike(queryContext.CriteriaBuilder, queryContext.Root.get(LanguageEntity._code), this.like));
        }
        if (this.isActives != null) {
            CriteriaBuilder.In<IsActive> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(LanguageEntity._isActive));
            for (IsActive item : this.isActives)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.codes != null) {
            CriteriaBuilder.In<String> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(LanguageEntity._code));
            for (String item : this.codes)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.excludedIds != null) {
            CriteriaBuilder.In<UUID> notInClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(LanguageEntity._id));
            for (UUID item : this.excludedIds)
                notInClause.value(item);
            predicates.add(notInClause.not());
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return null;
        }
    }

    @Override
    protected LanguageEntity convert(Tuple tuple, Set<String> columns) {
        LanguageEntity item = new LanguageEntity();
        item.setId(QueryBase.convertSafe(tuple, columns, LanguageEntity._id, UUID.class));
        item.setTenantId(QueryBase.convertSafe(tuple, columns, LanguageEntity._tenantId, UUID.class));
        item.setCode(QueryBase.convertSafe(tuple, columns, LanguageEntity._code, String.class));
        item.setPayload(QueryBase.convertSafe(tuple, columns, LanguageEntity._payload, String.class));
        item.setOrdinal(QueryBase.convertSafe(tuple, columns, LanguageEntity._ordinal, Integer.class));
        item.setCreatedAt(QueryBase.convertSafe(tuple, columns, LanguageEntity._createdAt, Instant.class));
        item.setUpdatedAt(QueryBase.convertSafe(tuple, columns, LanguageEntity._updatedAt, Instant.class));
        item.setIsActive(QueryBase.convertSafe(tuple, columns, LanguageEntity._isActive, IsActive.class));
        return item;
    }

    @Override
    protected String fieldNameOf(FieldResolver item) {
        if (item.match(Language._id)) return LanguageEntity._id;
        else if (item.match(Language._code)) return LanguageEntity._code;
        else if (item.match(Language._payload)) return LanguageEntity._payload;
        else if (item.match(Language._ordinal)) return LanguageEntity._ordinal;
        else if (item.match(Language._createdAt)) return LanguageEntity._createdAt;
        else if (item.match(Language._updatedAt)) return LanguageEntity._updatedAt;
        else if (item.match(Language._hash)) return LanguageEntity._updatedAt;
        else if (item.match(Language._isActive)) return LanguageEntity._isActive;
        else if (item.match(Language._belongsToCurrentTenant)) return LanguageEntity._tenantId;
        else return null;
    }

}
