package eu.eudat.query;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.authorization.Permission;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.commons.scope.user.UserScope;
import eu.eudat.data.DescriptionTagEntity;
import eu.eudat.data.DmpReferenceEntity;
import eu.eudat.data.ReferenceEntity;
import eu.eudat.data.TagEntity;
import eu.eudat.model.Tag;
import eu.eudat.query.utils.BuildSubQueryInput;
import eu.eudat.query.utils.QueryUtilsService;
import gr.cite.commons.web.authz.service.AuthorizationService;
import gr.cite.tools.data.query.FieldResolver;
import gr.cite.tools.data.query.QueryBase;
import gr.cite.tools.data.query.QueryContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.Predicate;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class TagQuery extends QueryBase<TagEntity> {

    private String like;
    private Collection<String> tags;
    private Collection<String> excludedTags;

    private Collection<UUID> ids;

    private Collection<UUID> excludedIds;

    private Collection<IsActive> isActives;

    private Collection<UUID> createdByIds;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);
    
    private final UserScope userScope;
    private final AuthorizationService authService;
    private final QueryUtilsService queryUtilsService;

    public TagQuery(UserScope userScope, AuthorizationService authService, QueryUtilsService queryUtilsService) {
	    this.userScope = userScope;
	    this.authService = authService;
	    this.queryUtilsService = queryUtilsService;
    }

    public TagQuery like(String value) {
        this.like = value;
        return this;
    }

    public TagQuery ids(UUID value) {
        this.ids = List.of(value);
        return this;
    }

    public TagQuery ids(UUID... value) {
        this.ids = Arrays.asList(value);
        return this;
    }

    public TagQuery ids(Collection<UUID> values) {
        this.ids = values;
        return this;
    }

    public TagQuery tags(String value) {
        this.tags = List.of(value);
        return this;
    }

    public TagQuery tags(String... value) {
        this.tags = Arrays.asList(value);
        return this;
    }

    public TagQuery tags(Collection<String> values) {
        this.tags = values;
        return this;
    }

    public TagQuery excludedTags(String value) {
        this.excludedTags = List.of(value);
        return this;
    }

    public TagQuery excludedTags(String... value) {
        this.excludedTags = Arrays.asList(value);
        return this;
    }

    public TagQuery excludedTags(Collection<String> values) {
        this.excludedTags = values;
        return this;
    }

    public TagQuery excludedIds(Collection<UUID> values) {
        this.excludedIds = values;
        return this;
    }

    public TagQuery excludedIds(UUID value) {
        this.excludedIds = List.of(value);
        return this;
    }

    public TagQuery excludedIds(UUID... value) {
        this.excludedIds = Arrays.asList(value);
        return this;
    }

    public TagQuery isActive(IsActive value) {
        this.isActives = List.of(value);
        return this;
    }

    public TagQuery isActive(IsActive... value) {
        this.isActives = Arrays.asList(value);
        return this;
    }

    public TagQuery isActive(Collection<IsActive> values) {
        this.isActives = values;
        return this;
    }
    
    public TagQuery createdByIds(UUID value) {
        this.createdByIds = List.of(value);
        return this;
    }

    public TagQuery createdByIds(UUID... value) {
        this.createdByIds = Arrays.asList(value);
        return this;
    }

    public TagQuery createdByIds(Collection<UUID> values) {
        this.createdByIds = values;
        return this;
    }

    public TagQuery authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    protected Boolean isFalseQuery() {
        return
                this.isEmpty(this.ids) ||
                        this.isEmpty(this.isActives) ||
                        this.isEmpty(this.tags) ||
                        this.isEmpty(this.excludedTags) ||
                        this.isEmpty(this.excludedIds) ||
                        this.isEmpty(this.createdByIds);
    }

    @Override
    protected Class<TagEntity> entityClass() {
        return TagEntity.class;
    }
    @Override
    protected <X, Y> Predicate applyAuthZ(QueryContext<X, Y> queryContext) {
        if (this.authorize.contains(AuthorizationFlags.None)) return null;
        if (this.authorize.contains(AuthorizationFlags.Permission) && this.authService.authorize(Permission.BrowseTag)) return null;
        UUID userId;
        boolean usePublic = this.authorize.contains(AuthorizationFlags.Public);
        if (this.authorize.contains(AuthorizationFlags.DmpAssociated)) userId = this.userScope.getUserIdSafe();
        else  userId = null;

        List<Predicate> predicates = new ArrayList<>();
        if (userId != null || usePublic ) {
            predicates.add(queryContext.CriteriaBuilder.or(
                    queryContext.CriteriaBuilder.isNull(queryContext.Root.get(TagEntity._createdById)),
                    userId != null ?  queryContext.CriteriaBuilder.equal(queryContext.Root.get(TagEntity._createdById), userId) : queryContext.CriteriaBuilder.or(),  //Creates a false query
                    queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._id)).value(queryUtilsService.buildSubQuery(new BuildSubQueryInput<>(new BuildSubQueryInput.Builder<>(DescriptionTagEntity.class, UUID.class)
                            .query(queryContext.Query)
                            .criteriaBuilder(queryContext.CriteriaBuilder)
                            .keyPathFunc((subQueryRoot) -> subQueryRoot.get(DescriptionTagEntity._tagId))
                            .filterFunc((subQueryRoot, cb) ->
                                    cb.in(subQueryRoot.get(DescriptionTagEntity._descriptionId)).value(queryUtilsService.buildDescriptionAuthZSubQuery(queryContext.Query, queryContext.CriteriaBuilder, userId, usePublic))
                            )
                    )))  //Creates a false query
            ));
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return queryContext.CriteriaBuilder.or(); //Creates a false query
        }
    }

    @Override
    protected <X, Y> Predicate applyFilters(QueryContext<X, Y> queryContext) {
        List<Predicate> predicates = new ArrayList<>();
        if (this.like != null && !this.like.isEmpty()) {
            predicates.add(queryUtilsService.ilike(queryContext.CriteriaBuilder, queryContext.Root.get(TagEntity._label), this.like));
        }
        if (this.ids != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._id));
            for (UUID item : this.ids)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.tags != null) {
            CriteriaBuilder.In<String> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._label));
            for (String item : this.tags)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.excludedTags != null) {
            CriteriaBuilder.In<String> notInClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._label));
            for (String item : this.excludedTags)
                notInClause.value(item);
            predicates.add(notInClause.not());
        }
        if (this.excludedIds != null) {
            CriteriaBuilder.In<UUID> notInClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._id));
            for (UUID item : this.excludedIds)
                notInClause.value(item);
            predicates.add(notInClause.not());
        }
        if (this.isActives != null) {
            CriteriaBuilder.In<IsActive> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._isActive));
            for (IsActive item : this.isActives)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.createdByIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(TagEntity._createdById));
            for (UUID item : this.createdByIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return null;
        }
    }

    @Override
    protected String fieldNameOf(FieldResolver item) {
        if (item.match(Tag._id)) return TagEntity._id;
        else if (item.match(Tag._label)) return TagEntity._label;
        else if (item.prefix(Tag._createdBy)) return TagEntity._createdById;
        else if (item.match(Tag._createdBy)) return TagEntity._createdById;
        else if (item.match(Tag._createdAt)) return TagEntity._createdAt;
        else if (item.match(Tag._updatedAt)) return TagEntity._updatedAt;
        else if (item.match(Tag._hash)) return TagEntity._updatedAt;
        else if (item.match(Tag._isActive)) return TagEntity._isActive;
        else if (item.match(Tag._belongsToCurrentTenant)) return TagEntity._tenantId;
        else return null;
    }

    @Override
    protected TagEntity convert(Tuple tuple, Set<String> columns) {
        TagEntity item = new TagEntity();
        item.setId(QueryBase.convertSafe(tuple, columns, TagEntity._id, UUID.class));
        item.setTenantId(QueryBase.convertSafe(tuple, columns, TagEntity._tenantId, UUID.class));
        item.setLabel(QueryBase.convertSafe(tuple, columns, TagEntity._label, String.class));
        item.setCreatedById(QueryBase.convertSafe(tuple, columns, TagEntity._createdById, UUID.class));
        item.setCreatedAt(QueryBase.convertSafe(tuple, columns, TagEntity._createdAt, Instant.class));
        item.setUpdatedAt(QueryBase.convertSafe(tuple, columns, TagEntity._updatedAt, Instant.class));
        item.setIsActive(QueryBase.convertSafe(tuple, columns, TagEntity._isActive, IsActive.class));
        return item;
    }

}
