package eu.eudat.service.maintenance;

import eu.eudat.authorization.Permission;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.data.TenantEntity;
import eu.eudat.data.TenantEntityManager;
import eu.eudat.data.UserEntity;
import eu.eudat.integrationevent.outbox.tenantremoval.TenantRemovalIntegrationEventHandler;
import eu.eudat.integrationevent.outbox.tenanttouched.TenantTouchedIntegrationEvent;
import eu.eudat.integrationevent.outbox.tenanttouched.TenantTouchedIntegrationEventHandler;
import eu.eudat.integrationevent.outbox.userremoval.UserRemovalIntegrationEventHandler;
import eu.eudat.integrationevent.outbox.usertouched.UserTouchedIntegrationEventHandler;
import eu.eudat.model.Tenant;
import eu.eudat.model.User;
import eu.eudat.query.TenantQuery;
import eu.eudat.query.UserQuery;
import gr.cite.commons.web.authz.service.AuthorizationService;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import javax.management.InvalidApplicationException;
import java.util.List;

@Service
public class MaintenanceServiceImpl implements MaintenanceService {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(MaintenanceServiceImpl.class));
    private final TenantEntityManager entityManager;
    private final AuthorizationService authorizationService;
    private final QueryFactory queryFactory;

    private final UserTouchedIntegrationEventHandler userTouchedIntegrationEventHandler;

    private final UserRemovalIntegrationEventHandler userRemovalIntegrationEventHandler;

    private final TenantTouchedIntegrationEventHandler tenantTouchedIntegrationEventHandler;

    private final TenantRemovalIntegrationEventHandler tenantRemovalIntegrationEventHandler;
    

    public MaintenanceServiceImpl(
		    TenantEntityManager entityManager, AuthorizationService authorizationService,
		    QueryFactory queryFactory, UserTouchedIntegrationEventHandler userTouchedIntegrationEventHandler, UserRemovalIntegrationEventHandler userRemovalIntegrationEventHandler, TenantTouchedIntegrationEventHandler tenantTouchedIntegrationEventHandler, TenantRemovalIntegrationEventHandler tenantRemovalIntegrationEventHandler) {
        this.entityManager = entityManager;
        this.authorizationService = authorizationService;
        this.queryFactory = queryFactory;
	    this.userTouchedIntegrationEventHandler = userTouchedIntegrationEventHandler;
	    this.userRemovalIntegrationEventHandler = userRemovalIntegrationEventHandler;
	    this.tenantTouchedIntegrationEventHandler = tenantTouchedIntegrationEventHandler;
	    this.tenantRemovalIntegrationEventHandler = tenantRemovalIntegrationEventHandler;
    }


    @Override
    public void sendUserTouchEvents() throws InvalidApplicationException {
        logger.debug("send user touch queue events");
        this.authorizationService.authorizeForce(Permission.ManageQueueEvents);

        List<UserEntity> activeUsers;
        List<UserEntity> inactiveUsers;
        try {
            this.entityManager.disableTenantFilters();
            UserQuery userQuery = queryFactory.query(UserQuery.class);
            userQuery.isActive(IsActive.Active);
            activeUsers = userQuery.collectAs(new BaseFieldSet().ensure(User._id));
            userQuery.isActive(IsActive.Inactive);
            inactiveUsers = userQuery.collectAs(new BaseFieldSet().ensure(User._id));
            this.entityManager.enableTenantFilters();

        } finally {
            this.entityManager.enableTenantFilters();
        }
        for(UserEntity user : activeUsers)
            this.userTouchedIntegrationEventHandler.handle(user.getId());

        for(UserEntity user : inactiveUsers)
            this.userRemovalIntegrationEventHandler.handle(user.getId());
    }
    
    @Override
    public void sendTenantTouchEvents() throws InvalidApplicationException {
        logger.debug("send tenant touch queue events");
        this.authorizationService.authorizeForce(Permission.ManageQueueEvents);

        List<TenantEntity> activeTenants;
        List<TenantEntity> inactiveTenants;
        try {
            this.entityManager.disableTenantFilters();
            TenantQuery tenantQuery = queryFactory.query(TenantQuery.class);
            tenantQuery.isActive(IsActive.Active);
            activeTenants = tenantQuery.collectAs(new BaseFieldSet().ensure(Tenant._id).ensure(Tenant._code));
            tenantQuery.isActive(IsActive.Inactive);
            inactiveTenants = tenantQuery.collectAs(new BaseFieldSet().ensure(Tenant._id).ensure(Tenant._code));
            
        } finally {
            this.entityManager.enableTenantFilters();
        }
        for (TenantEntity tenant : activeTenants) {
            TenantTouchedIntegrationEvent event = new TenantTouchedIntegrationEvent();
            event.setId(tenant.getId());
            event.setCode(tenant.getCode());
            this.tenantTouchedIntegrationEventHandler.handle(event);
        }

        for (TenantEntity tenant : inactiveTenants) {
            this.tenantRemovalIntegrationEventHandler.handle(tenant.getId());
        }
    }

}
