package eu.eudat.controllers;

import eu.eudat.audit.AuditableAction;
import eu.eudat.authorization.AuthorizationFlags;
import gr.cite.tools.validation.ValidationFilterAnnotation;
import eu.eudat.data.UserSettingsEntity;
import eu.eudat.model.UserSettings;
import eu.eudat.model.builder.UserSettingsBuilder;
import eu.eudat.model.censorship.UserSettingsCensor;
import eu.eudat.model.persist.UserSettingsPersist;
import eu.eudat.model.result.QueryResult;
import eu.eudat.query.UserSettingsQuery;
import eu.eudat.query.lookup.UserSettingsLookup;
import eu.eudat.service.user.settings.UserSettingsService;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.censor.CensorFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.exception.MyForbiddenException;
import gr.cite.tools.exception.MyNotFoundException;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.logging.MapLogEntry;
import jakarta.transaction.Transactional;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.web.bind.annotation.*;

import javax.management.InvalidApplicationException;
import java.util.AbstractMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

@RestController
@RequestMapping(path = "api/user-settings")
public class UserSettingsController {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(UserSettingsController.class));

    private final BuilderFactory builderFactory;

    private final AuditService auditService;

    private final UserSettingsService settingsService;

    private final CensorFactory censorFactory;

    private final QueryFactory queryFactory;

    private final MessageSource messageSource;

    @Autowired
    public UserSettingsController(
            BuilderFactory builderFactory,
            AuditService auditService,
            UserSettingsService settingsService,
            CensorFactory censorFactory,
            QueryFactory queryFactory,
            MessageSource messageSource) {
        this.builderFactory = builderFactory;
        this.auditService = auditService;
        this.settingsService = settingsService;
        this.censorFactory = censorFactory;
        this.queryFactory = queryFactory;
        this.messageSource = messageSource;
    }

    @PostMapping("query")
    public QueryResult<UserSettings> Query(@RequestBody UserSettingsLookup lookup) throws MyApplicationException, MyForbiddenException {
        logger.debug("querying {}", UserSettings.class.getSimpleName());
        this.censorFactory.censor(UserSettingsCensor.class).censor(lookup.getProject(), null);
        UserSettingsQuery query = lookup.enrich(this.queryFactory).authorize(AuthorizationFlags.OwnerOrDmpAssociatedOrPermission);
        List<UserSettingsEntity> data = query.collectAs(lookup.getProject());
        List<UserSettings> models = this.builderFactory.builder(UserSettingsBuilder.class).authorize(AuthorizationFlags.OwnerOrDmpAssociatedOrPermission).build(lookup.getProject(), data);
        long count = (lookup.getMetadata() != null && lookup.getMetadata().getCountAll()) ? query.count() : models.size();

        this.auditService.track(AuditableAction.User_Settings_Query, "lookup", lookup);

        return new QueryResult<>(models, count);
    }

    @GetMapping("{key}")
    @Transactional
    public UserSettings Get(@PathVariable("key") String key) throws MyApplicationException, MyForbiddenException, MyNotFoundException {
        logger.debug(new MapLogEntry("retrieving" + UserSettings.class.getSimpleName()).And("key", key));

        BaseFieldSet fieldSet = new BaseFieldSet();
        fieldSet.setFields(Set.of(
                UserSettings._id,
                UserSettings._key,
                UserSettings._value,
                UserSettings._entityId,
                UserSettings._createdAt,
                UserSettings._updatedAt,
                UserSettings._type
        ));
        UserSettingsQuery query = this.queryFactory.query(UserSettingsQuery.class).authorize(AuthorizationFlags.OwnerOrDmpAssociatedOrPermission).keys(key);
        UserSettings model = this.builderFactory.builder(UserSettingsBuilder.class).authorize(AuthorizationFlags.OwnerOrDmpAssociatedOrPermission).build(fieldSet, query.firstAs(fieldSet));

        this.auditService.track(AuditableAction.User_Settings_Lookup, Map.ofEntries(
                new AbstractMap.SimpleEntry<String, Object>("key", key)
        ));

        return model;
    }

    @PostMapping("persist")
    @Transactional
    @ValidationFilterAnnotation(validator = UserSettingsPersist.UserSettingsPersistValidator.ValidatorName, argumentName = "model")
    public UserSettings Persist(@RequestBody UserSettingsPersist model, FieldSet fieldSet) throws MyApplicationException, MyForbiddenException, MyNotFoundException, InvalidApplicationException {
        logger.debug(new MapLogEntry("persisting" + UserSettings.class.getSimpleName()).And("model", model).And("fieldSet", fieldSet));

        UserSettings persisted = this.settingsService.persist(model, fieldSet);

        this.auditService.track(AuditableAction.User_Settings_Persist, Map.ofEntries(
                new AbstractMap.SimpleEntry<String, Object>("model", model),
                new AbstractMap.SimpleEntry<String, Object>("fields", fieldSet)
        ));

        return persisted;
    }

}
