import { HttpClient, HttpHeaders, HttpParamsOptions, HttpResponse } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { DmpStatus } from '@app/core/common/enum/dmp-status';
import { DmpUserRole } from '@app/core/common/enum/dmp-user-role';
import { IsActive } from '@app/core/common/enum/is-active.enum';
import { DmpDescriptionTemplateLookup } from '@app/core/query/dmp-description-template.lookup';
import { DmpLookup } from '@app/core/query/dmp.lookup';
import { MultipleAutoCompleteConfiguration } from '@app/library/auto-complete/multiple/multiple-auto-complete-configuration';
import { SingleAutoCompleteConfiguration } from '@app/library/auto-complete/single/single-auto-complete-configuration';
import { QueryResult } from '@common/model/query-result';
import { FilterService } from '@common/modules/text-filter/filter-service';
import { Guid } from '@common/types/guid';
import { Observable, throwError } from 'rxjs';
import { catchError, map } from 'rxjs/operators';
import { nameof } from 'ts-simple-nameof';
import { BaseHttpParams } from '../../../../common/http/base-http-params';
import { InterceptorType } from '../../../../common/http/interceptors/interceptor-type';
import { CloneDmpPersist, Dmp, DmpPersist, DmpUser, DmpUserInvitePersist, DmpUserPersist, DmpUserRemovePersist, NewVersionDmpPersist, PublicDmp } from '../../model/dmp/dmp';
import { AuthService } from '../auth/auth.service';
import { ConfigurationService } from '../configuration/configuration.service';
import { BaseHttpV2Service } from '../http/base-http-v2.service';
import { DmpValidationResult } from '@app/ui/dmp/dmp-finalize-dialog/dmp-finalize-dialog.component';

@Injectable()
export class DmpService {

	private headers = new HttpHeaders();

	constructor(
		private http: BaseHttpV2Service,
		private httpClient: HttpClient,
		private configurationService: ConfigurationService,
		private filterService: FilterService,
		private authService: AuthService
	) {
	}

	private get apiBase(): string { return `${this.configurationService.server}dmp`; }

	query(q: DmpLookup): Observable<QueryResult<Dmp>> {
		const url = `${this.apiBase}/query`;
		return this.http.post<QueryResult<Dmp>>(url, q).pipe(catchError((error: any) => throwError(error)));
	}

	publicQuery(q: DmpLookup): Observable<QueryResult<PublicDmp>> {
		const url = `${this.apiBase}/public/query`;
		const params = new BaseHttpParams();
		params.interceptorContext = {
			excludedInterceptors: [InterceptorType.AuthToken,
				InterceptorType.TenantHeaderInterceptor]
		};
		return this.http.post<QueryResult<PublicDmp>>(url, q, {params: params}).pipe(catchError((error: any) => throwError(error)));
	}

	getSingle(id: Guid, reqFields: string[] = []): Observable<Dmp> {
		const url = `${this.apiBase}/${id}`;
		const options = { params: { f: reqFields } };

		return this.http
			.get<Dmp>(url, options).pipe(
				catchError((error: any) => throwError(error)));
	}

	getPublicSingle(id: Guid, reqFields: string[] = []): Observable<PublicDmp> {
		const url = `${this.apiBase}/public/${id}`;

		const options: HttpParamsOptions = { fromObject: { f: reqFields } };

		let params: BaseHttpParams = new BaseHttpParams(options);
		params.interceptorContext = {
			excludedInterceptors: [InterceptorType.AuthToken,
				InterceptorType.TenantHeaderInterceptor]
		};

		return this.http
			.get<PublicDmp>(url, { params: params }).pipe(
				catchError((error: any) => throwError(error)));
	}

	persist(item: DmpPersist): Observable<Dmp> {
		const url = `${this.apiBase}/persist`;

		return this.http
			.post<Dmp>(url, item).pipe(
				catchError((error: any) => throwError(error)));
	}

	delete(id: Guid): Observable<Dmp> {
		const url = `${this.apiBase}/${id}`;

		return this.http
			.delete<Dmp>(url).pipe(
				catchError((error: any) => throwError(error)));
	}

	finalize(id: Guid, descriptionIds: Guid[] = []): Observable<Boolean> {
		const url = `${this.apiBase}/finalize/${id}`;

		return this.http
			.post<Boolean>(url, {descriptionIds: descriptionIds}).pipe(
				catchError((error: any) => throwError(error)));
	}

	undoFinalize(id: Guid, reqFields: string[] = []): Observable<Boolean> {
		const url = `${this.apiBase}/undo-finalize/${id}`;

		const options = { params: { f: reqFields } };

		return this.http
			.get<Boolean>(url, options).pipe(
				catchError((error: any) => throwError(error)));
	}

	validate(id: Guid): Observable<DmpValidationResult> {
		const url = `${this.apiBase}/validate/${id}`;

		return this.http
			.get<DmpValidationResult>(url).pipe(
				catchError((error: any) => throwError(error)));
	}

	clone(item: CloneDmpPersist, reqFields: string[] = []): Observable<Dmp> {
		const url = `${this.apiBase}/clone`;
		const options = { params: { f: reqFields } };

		return this.http
			.post<Dmp>(url, item, options).pipe(
				catchError((error: any) => throwError(error)));
	}

	newVersion(item: NewVersionDmpPersist, reqFields: string[] = []): Observable<Dmp> {
		const url = `${this.apiBase}/new-version`;
		const options = { params: { f: reqFields } };

		return this.http
			.post<Dmp>(url, item, options ).pipe(
				catchError((error: any) => throwError(error)));
	}

	assignUsers(id: Guid, items: DmpUserPersist[], reqFields: string[] = []): Observable<DmpUser> {
		const url = `${this.apiBase}/${id}/assign-users`;
		const options = { params: { f: reqFields } };

		return this.http
			.post<DmpUser>(url, items).pipe(
				catchError((error: any) => throwError(error)));
	}

	removeUser(item: DmpUserRemovePersist, reqFields: string[] = []): Observable<Dmp> {
		const url = `${this.apiBase}/remove-user`;

		return this.http
			.post<Dmp>(url, item).pipe(
				catchError((error: any) => throwError(error)));
	}

	inviteUsers(dmpId: Guid, item: DmpUserInvitePersist): Observable<any> {
		const url = `${this.apiBase}/${dmpId}/invite-users`;

		return this.http
			.post<any>(url, item).pipe(
				catchError((error: any) => throwError(error)));
	}

	acceptInvitation(dmpId: Guid, token: string): Observable<any> {
		const url = `${this.apiBase}/${dmpId}/token/${token}/invite-accept`;

		return this.http.get<any>(url).pipe(catchError((error: any) => throwError(error)));
	}

	public download(id: string, format: string): Observable<HttpResponse<Blob>> {
		//let headerDoc: HttpHeaders = this.headers.set('Content-Type', 'application/msword')
		return this.httpClient.get(`${this.apiBase}/${id}/export/${format}`, { responseType: 'blob', observe: 'response', headers: this.headers });
	}

	public downloadPDF(id: string): Observable<HttpResponse<Blob>> {
		let headerPdf: HttpHeaders = this.headers.set('Content-Type', 'application/pdf')
		return this.httpClient.get(`${this.apiBase}/${id}/export/Pdf`, { responseType: 'blob', observe: 'response', headers: headerPdf });
	}

	// public downloadJson(id: string): Observable<HttpResponse<Blob>> {
	// 	return this.httpClient.get(this.actionUrl + 'rda/' + id, { responseType: 'blob', observe: 'response' });
	// }

	downloadXML(id: Guid): Observable<HttpResponse<Blob>> {
		const url = `${this.apiBase}/xml/export/${id}`;
		let headerXml: HttpHeaders = this.headers.set('Content-Type', 'application/xml');
		const params = new BaseHttpParams();
		params.interceptorContext = {
			excludedInterceptors: [InterceptorType.JSONContentType]
		};
		return this.httpClient.get(url, { params: params, responseType: 'blob', observe: 'response', headers: headerXml });
	}

	uploadFile(file: FileList, labelSent: string, reqFields: string[] = []): Observable<Dmp> {
		const url = `${this.apiBase}/xml/import`;
		const params = new BaseHttpParams();
		params.interceptorContext = {
			excludedInterceptors: [InterceptorType.JSONContentType]
		};
		const formData = new FormData();
		formData.append('file', file[0], labelSent);
		return this.http.post(url, formData, { params: params });
	}

	//
	// Autocomplete Commons
	//
	// tslint:disable-next-line: member-ordering
	singleAutocompleteConfiguration: SingleAutoCompleteConfiguration = {
		initialItems: (data?: any) => this.query(this.buildAutocompleteLookup()).pipe(map(x => x.items)),
		filterFn: (searchQuery: string, data?: any) => this.query(this.buildAutocompleteLookup(searchQuery)).pipe(map(x => x.items)),
		getSelectedItem: (selectedItem: any) => this.query(this.buildAutocompleteLookup(null, null, [selectedItem])).pipe(map(x => x.items[0])),
		displayFn: (item: Dmp) => item.label,
		titleFn: (item: Dmp) => item.label,
		valueAssign: (item: Dmp) => item.id,
	};

	// tslint:disable-next-line: member-ordering
	multipleAutocompleteConfiguration: MultipleAutoCompleteConfiguration = {
		initialItems: (excludedItems: any[], data?: any) => this.query(this.buildAutocompleteLookup(null, excludedItems ? excludedItems : null)).pipe(map(x => x.items)),
		filterFn: (searchQuery: string, excludedItems: any[]) => this.query(this.buildAutocompleteLookup(searchQuery, excludedItems)).pipe(map(x => x.items)),
		getSelectedItems: (selectedItems: any[]) => this.query(this.buildAutocompleteLookup(null, null, selectedItems)).pipe(map(x => x.items)),
		displayFn: (item: Dmp) => item.label,
		titleFn: (item: Dmp) => item.label,
		valueAssign: (item: Dmp) => item.id,
	};

	public buildAutocompleteLookup(like?: string, excludedIds?: Guid[], ids?: Guid[], statuses?: DmpStatus[], dmpDescriptionTemplateSubQuery?: DmpDescriptionTemplateLookup): DmpLookup {
		const lookup: DmpLookup = new DmpLookup();
		lookup.page = { size: 100, offset: 0 };
		if (excludedIds && excludedIds.length > 0) { lookup.excludedIds = excludedIds; }
		if (ids && ids.length > 0) { lookup.ids = ids; }
		lookup.isActive = [IsActive.Active];
		lookup.statuses = statuses;
		lookup.project = {
			fields: [
				nameof<Dmp>(x => x.id),
				nameof<Dmp>(x => x.label)
			]
		};
		if (dmpDescriptionTemplateSubQuery != null) lookup.dmpDescriptionTemplateSubQuery = dmpDescriptionTemplateSubQuery;
		lookup.order = { items: [nameof<Dmp>(x => x.label)] };
		if (like) { lookup.like = this.filterService.transformLike(like); }
		return lookup;
	}

	//
	//
	// UI Helpers
	//
	//

	getCurrentUserRolesInDmp(dmpUsers: DmpUser[]): DmpUserRole[] {
		const principalId: Guid = this.authService.userId();
		let dmpUserRoles: DmpUserRole[] = null;
		if (principalId) {
			dmpUserRoles = dmpUsers.filter(element => element.isActive == IsActive.Active && element?.user?.id === principalId).map(x => x.role);
		}
		return dmpUserRoles;
	}
}
