
import { Component, OnInit, TemplateRef, ViewChild } from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { DmpBlueprintStatus } from '@app/core/common/enum/dmp-blueprint-status';
import { IsActive } from '@app/core/common/enum/is-active.enum';
import { DmpBlueprint } from '@app/core/model/dmp-blueprint/dmp-blueprint';
import { DmpBlueprintLookup } from '@app/core/query/dmp-blueprint.lookup';
import { AuthService } from '@app/core/services/auth/auth.service';
import { DmpBlueprintService } from '@app/core/services/dmp/dmp-blueprint.service';
import { MatomoService } from '@app/core/services/matomo/matomo-service';
import { SnackBarNotificationLevel, UiNotificationService } from '@app/core/services/notification/ui-notification-service';
import { EnumUtils } from '@app/core/services/utilities/enum-utils.service';
import { FileUtils } from '@app/core/services/utilities/file-utils.service';
import { QueryParamsService } from '@app/core/services/utilities/query-params.service';
import { BaseListingComponent } from '@common/base/base-listing-component';
import { PipeService } from '@common/formatting/pipe.service';
import { DataTableDateTimeFormatPipe } from '@common/formatting/pipes/date-time-format.pipe';
import { QueryResult } from '@common/model/query-result';
import { ConfirmationDialogComponent } from '@common/modules/confirmation-dialog/confirmation-dialog.component';
import { HttpErrorHandlingService } from '@common/modules/errors/error-handling/http-error-handling.service';
import { ColumnDefinition, ColumnsChangedEvent, HybridListingComponent, PageLoadEvent } from '@common/modules/hybrid-listing/hybrid-listing.component';
import { Guid } from '@common/types/guid';
import { TranslateService } from '@ngx-translate/core';
import * as FileSaver from 'file-saver';
import { Observable } from 'rxjs';
import { takeUntil } from 'rxjs/operators';
import { nameof } from 'ts-simple-nameof';
import { ImportDmpBlueprintDialogComponent } from './import-dmp-blueprint/import-dmp-blueprint.dialog.component';
import { IsActiveTypePipe } from '@common/formatting/pipes/is-active-type.pipe';
import { DmpBlueprintVersionStatus } from '@app/core/common/enum/dmp-blueprint-version-status';


@Component({
	selector: 'app-dmp-blueprint-listing-component',
	templateUrl: './dmp-blueprint-listing.component.html',
	styleUrls: ['./dmp-blueprint-listing.component.scss']
})
export class DmpBlueprintListingComponent extends BaseListingComponent<DmpBlueprint, DmpBlueprintLookup> implements OnInit {
	publish = false;
	userSettingsKey = { key: 'DmpBlueprintListingUserSettings' };
	propertiesAvailableForOrder: ColumnDefinition[];
	dmpBlueprintStatuses = DmpBlueprintStatus;
	mode;

	@ViewChild('dmpBlueprintStatus', { static: true }) dmpBlueprintStatus?: TemplateRef<any>;
	@ViewChild('actions', { static: true }) actions?: TemplateRef<any>;
	@ViewChild(HybridListingComponent, { static: true }) hybridListingComponent: HybridListingComponent;

	private readonly lookupFields: string[] = [
		nameof<DmpBlueprint>(x => x.id),
		nameof<DmpBlueprint>(x => x.label),
		nameof<DmpBlueprint>(x => x.status),
		nameof<DmpBlueprint>(x => x.version),
		nameof<DmpBlueprint>(x => x.groupId),
		nameof<DmpBlueprint>(x => x.updatedAt),
		nameof<DmpBlueprint>(x => x.createdAt),
		nameof<DmpBlueprint>(x => x.hash),
		nameof<DmpBlueprint>(x => x.isActive),
		nameof<DmpBlueprint>(x => x.belongsToCurrentTenant)
	];

	rowIdentity = x => x.id;

	constructor(
		protected router: Router,
		protected route: ActivatedRoute,
		protected uiNotificationService: UiNotificationService,
		protected httpErrorHandlingService: HttpErrorHandlingService,
		protected queryParamsService: QueryParamsService,
		private dmpBlueprintService: DmpBlueprintService,
		public authService: AuthService,
		private matomoService: MatomoService,
		private pipeService: PipeService,
		public enumUtils: EnumUtils,
		private language: TranslateService,
		private dialog: MatDialog,
		private fileUtils: FileUtils
	) {
		super(router, route, uiNotificationService, httpErrorHandlingService, queryParamsService);
		// Lookup setup
		// Default lookup values are defined in the user settings class.
		this.mode = this.route.snapshot?.data['mode'];
		this.lookup = this.initializeLookup();
	}

	ngOnInit() {
		this.matomoService.trackPageView('Admin: DMP Templates');
		super.ngOnInit();
	}

	protected initializeLookup(): DmpBlueprintLookup {
		const lookup = new DmpBlueprintLookup();
		lookup.metadata = { countAll: true };
		lookup.page = { offset: 0, size: this.ITEMS_PER_PAGE };
		lookup.isActive = [IsActive.Active];
		lookup.order = { items: [this.toDescSortField(nameof<DmpBlueprint>(x => x.createdAt))] };
		if (this.mode && this.mode == 'versions-listing') {
			lookup.groupIds = [Guid.parse(this.route.snapshot.paramMap.get('groupid'))];
			lookup.versionStatuses = null;
		}else{
			lookup.versionStatuses = [DmpBlueprintVersionStatus.Current, DmpBlueprintVersionStatus.NotFinalized];
		}
		this.updateOrderUiFields(lookup.order);

		lookup.project = {
			fields: this.lookupFields
		};

		return lookup;
	}

	protected setupColumns() {
		this.gridColumns.push(...[{
			prop: nameof<DmpBlueprint>(x => x.label),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.NAME'
		},
		{
			prop: nameof<DmpBlueprint>(x => x.status),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.STATUS',
			cellTemplate: this.dmpBlueprintStatus
		},
		{
			prop: nameof<DmpBlueprint>(x => x.version),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.VERSION'
		},
		{
			prop: nameof<DmpBlueprint>(x => x.createdAt),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.CREATED-AT',
			pipe: this.pipeService.getPipe<DataTableDateTimeFormatPipe>(DataTableDateTimeFormatPipe).withFormat('short')
		},
		{
			prop: nameof<DmpBlueprint>(x => x.updatedAt),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.UPDATED-AT',
			pipe: this.pipeService.getPipe<DataTableDateTimeFormatPipe>(DataTableDateTimeFormatPipe).withFormat('short')
		},
		{
			prop: nameof<DmpBlueprint>(x => x.isActive),
			sortable: true,
			languageName: 'DMP-BLUEPRINT-LISTING.FIELDS.IS-ACTIVE',
			pipe: this.pipeService.getPipe<IsActiveTypePipe>(IsActiveTypePipe)
		},
		{
			alwaysShown: true,
			cellTemplate: this.actions,
			maxWidth: 120
		}
		]);
		this.propertiesAvailableForOrder = this.gridColumns.filter(x => x.sortable);
	}

	//
	// Listing Component functions
	//
	onColumnsChanged(event: ColumnsChangedEvent) {
		super.onColumnsChanged(event);
		this.onColumnsChangedInternal(event.properties.map(x => x.toString()));
	}

	private onColumnsChangedInternal(columns: string[]) {
		// Here are defined the projection fields that always requested from the api.
		const fields = new Set(this.lookupFields);
		this.gridColumns.map(x => x.prop)
			.filter(x => !columns?.includes(x as string))
			.forEach(item => {
				fields.delete(item as string)
			});
		this.lookup.project = { fields: [...fields] };
		this.onPageLoad({ offset: 0 } as PageLoadEvent);
	}

	protected loadListing(): Observable<QueryResult<DmpBlueprint>> {
		return this.dmpBlueprintService.query(this.lookup);
	}

	public delete(id: Guid) {
		if (id) {
			const dialogRef = this.dialog.open(ConfirmationDialogComponent, {
				data: {
					isDeleteConfirmation: true,
					message: this.language.instant('GENERAL.CONFIRMATION-DIALOG.DELETE-ITEM'),
					confirmButton: this.language.instant('GENERAL.CONFIRMATION-DIALOG.ACTIONS.CONFIRM'),
					cancelButton: this.language.instant('GENERAL.CONFIRMATION-DIALOG.ACTIONS.CANCEL')
				}
			});
			dialogRef.afterClosed().pipe(takeUntil(this._destroyed)).subscribe(result => {
				if (result) {
					this.dmpBlueprintService.delete(id).pipe(takeUntil(this._destroyed))
						.subscribe(
							complete => this.onCallbackSuccess(),
							error => this.onCallbackError(error)
						);
				}
			});
		}
	}

	onCallbackSuccess(): void {
		this.uiNotificationService.snackBarNotification(this.language.instant('GENERAL.SNACK-BAR.SUCCESSFUL-DELETE'), SnackBarNotificationLevel.Success);
		this.refresh();
	}

	export(id: Guid): void {
		this.dmpBlueprintService.downloadXML(id)
			.pipe(takeUntil(this._destroyed))
			.subscribe(response => {
				const blob = new Blob([response.body], { type: 'application/xml' });
				const filename = this.fileUtils.getFilenameFromContentDispositionHeader(response.headers.get('Content-Disposition'));
				FileSaver.saveAs(blob, filename);
			});
	}

	import(): void {
		const dialogRef = this.dialog.open(ImportDmpBlueprintDialogComponent, {
			restoreFocus: false,
			data: {
				message: this.language.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML-FILE-TITLE'),
				confirmButton: this.language.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML'),
				cancelButton: this.language.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML-FILE-CANCEL'),
				name: '',
				file: FileList,
				sucsess: false
			}
		});
		dialogRef.afterClosed().pipe(takeUntil(this._destroyed)).subscribe(data => {
			if (data && data.sucsess && data.name != null && data.file != null) {
				this.dmpBlueprintService.uploadFile(data.file, data.name)
					.pipe(takeUntil(this._destroyed))
					.subscribe(_ => {
						this.uiNotificationService.snackBarNotification(this.language.instant('DMP-BLUEPRINT-LISTING.MESSAGES.TEMPLATE-UPLOAD-SUCCESS'), SnackBarNotificationLevel.Success);
						this.refresh();
					},
						error => {
							this.uiNotificationService.snackBarNotification(error.message, SnackBarNotificationLevel.Error);
						});
			}
		});
	}
}


// ngOnInit() {


// 	refresh() {
// 		this.dataSource = new DatasetDataSource(this.dmpBlueprintService, this._paginator, this.sort, this.criteria);
// 	}

// 	clone(id: string) {
// 		this.router.navigate(['dmp-blueprints/clone/' + id]);
// 	}

// 	rowClick(rowId: String) {
// 		this.router.navigate(['dmp-blueprints/' + rowId]);
// 	}

// 	downloadXML(dmpBlueprintId: string): void {
// 		this.dmpBlueprintService.downloadXML(dmpBlueprintId)
// 			.pipe(takeUntil(this._destroyed))
// 			.subscribe(response => {
// 				const blob = new Blob([response.body], { type: 'application/xml' });
// 				const filename = this.fileUtils.getFilenameFromContentDispositionHeader(response.headers.get('Content-Disposition'));
// 				FileSaver.saveAs(blob, filename);
// 			});
// 	}

// 	deleteTemplate(id: string) {
// 		if (id) {
// 			this.dialog.open(ConfirmationDialogComponent, {
// 				data: {
// 					isDeleteConfirmation: true,
// 					confirmButton: this.languageService.instant('DMP-BLUEPRINT-EDITOR.CONFIRM-DELETE-DIALOG.CONFIRM-BUTTON'),
// 					cancelButton: this.languageService.instant("DMP-BLUEPRINT-EDITOR.CONFIRM-DELETE-DIALOG.CANCEL-BUTTON"),
// 					message: this.languageService.instant("DMP-BLUEPRINT-EDITOR.CONFIRM-DELETE-DIALOG.MESSAGE")
// 				}
// 			})
// 				.afterClosed()
// 				.subscribe(
// 					confirmed => {
// 						if (confirmed) {
// 							this.dmpBlueprintService.delete(id)
// 								.pipe(takeUntil(this._destroyed))
// 								.subscribe(
// 									complete => {
// 										this.uiNotificationService.snackBarNotification(this.languageService.instant('GENERAL.SNACK-BAR.SUCCESSFUL-DMP-BLUEPRINT-DELETE'), SnackBarNotificationLevel.Success);
// 										this.refresh();
// 									},
// 									error => {
// 										this.onCallbackError(error);
// 										if (error.error.statusCode == 674) {
// 											this.uiNotificationService.snackBarNotification(this.languageService.instant('GENERAL.SNACK-BAR.UNSUCCESSFUL-DMP-BLUEPRINT-DELETE'), SnackBarNotificationLevel.Error);
// 										} else {
// 											this.uiNotificationService.snackBarNotification(this.languageService.instant(error.message), SnackBarNotificationLevel.Error);
// 										}
// 									}
// 								);
// 						}
// 					}
// 				)
// 		}
// 	}

// 	onCallbackError(errorResponse: HttpErrorResponse) {
// 		this.uiNotificationService.snackBarNotification(errorResponse.message, SnackBarNotificationLevel.Warning);
// 	}

// 	getDefaultCriteria(): DmpBlueprintCriteria {
// 		const defaultCriteria = new DmpBlueprintCriteria();
// 		return defaultCriteria;
// 	}

// 	// makeItPublic(id: String) {
// 	//   debugger;
// 	//   this.datasetService.makeDatasetPublic(id).pipe(takeUntil(this._destroyed)).subscribe();
// 	// }

// 	parseStatus(value: number): string {
// 		const stringVal = value.toString()
// 		try {
// 			return this.statuses.find(status => status.value === stringVal).viewValue;
// 		} catch {
// 			return stringVal;
// 		}
// 	}

// 	openDialog(): void {
// 		const dialogRef = this.dialog.open(DialodConfirmationUploadDmpBlueprints, {
// 			restoreFocus: false,
// 			data: {
// 				message: this.languageService.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML-FILE-TITLE'),
// 				confirmButton: this.languageService.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML'),
// 				cancelButton: this.languageService.instant('DMP-BLUEPRINT-LISTING.IMPORT.UPLOAD-XML-FILE-CANCEL'),
// 				name: '',
// 				file: FileList,
// 				sucsess: false
// 			}
// 		});
// 		dialogRef.afterClosed().pipe(takeUntil(this._destroyed)).subscribe(data => {
// 			if (data && data.sucsess && data.name != null && data.file != null) {
// 				this.dmpBlueprintService.uploadFile(data.file, data.name)
// 					.pipe(takeUntil(this._destroyed))
// 					.subscribe(_ => {
// 						this.uiNotificationService.snackBarNotification(this.languageService.instant('DMP-BLUEPRINT-LISTING.MESSAGES.TEMPLATE-UPLOAD-SUCCESS'), SnackBarNotificationLevel.Success);
// 						this.refresh();
// 					},
// 						error => {
// 							this.uiNotificationService.snackBarNotification(error.message, SnackBarNotificationLevel.Error);
// 						});
// 			}
// 		});
// 	}
// 	getStatusClass(status: number): string {

// 		if (status === 1) {//finalized
// 			return 'status-chip-finalized'
// 		}
// 		if (status === 0) {
// 			return 'status-chip-draft';
// 		}
// 		return '';
// 	}
// }

// export class DatasetDataSource extends DataSource<DmpBlueprintListing> {

// 	totalCount = 0;

// 	constructor(
// 		private _service: dmpBlueprintService,
// 		private _paginator: MatPaginator,
// 		private _sort: MatSort,
// 		private _criteria: DmpBlueprintCriteriaComponent
// 	) {
// 		super();

// 	}

// 	connect(): Observable<DmpBlueprintListing[]> {
// 		const displayDataChanges = [
// 			this._paginator.page
// 			//this._sort.matSortChange
// 		];

// 		return observableMerge(...displayDataChanges).pipe(
// 			startWith(null),
// 			switchMap(() => {
// 				const startIndex = this._paginator.pageIndex * this._paginator.pageSize;
// 				let fields: Array<string> = new Array();
// 				if (this._sort.active) { fields = this._sort.direction === 'asc' ? ['+' + this._sort.active] : ['-' + this._sort.active]; }
// 				const request = new DataTableRequest<DmpBlueprintCriteria>(startIndex, this._paginator.pageSize, { fields: fields });
// 				request.criteria = this._criteria.criteria;
// 				return this._service.getPagedBlueprint(request);
// 			}),
// 			/*.catch((error: any) => {
// 			  this._snackBar.openFromComponent(SnackBarNotificationComponent, {
// 				data: { message: 'GENERAL.SNACK-BAR.FORMS-BAD-REQUEST', language: this._languageService },
// 				duration: 3000,
// 				extraClasses: ['snackbar-warning']
// 			  });
// 			  //this._criteria.criteria.onCallbackError(error);
// 			  return Observable.of(null);
// 			})*/
// 			map(result => {
// 				return result;
// 			}),
// 			map(result => {
// 				if (!result) { return []; }
// 				if (this._paginator.pageIndex === 0) { this.totalCount = result.totalCount; }
// 				return result.data;
// 			}));
// 	}

// 	disconnect() {
// 		// No-op
// 	}
