import { Location } from '@angular/common';
import { Component, Input, OnInit } from '@angular/core';
import { UntypedFormBuilder, UntypedFormControl } from '@angular/forms';
import { ActivatedRoute, Router } from '@angular/router';
import { RecentActivityOrder } from '@app/core/common/enum/recent-activity-order';
import { DescriptionTemplate } from '@app/core/model/description-template/description-template';
import { Description } from '@app/core/model/description/description';
import { RecentActivityItem } from '@app/core/model/dashboard/recent-activity-item';
import { Dmp, DmpUser } from '@app/core/model/dmp/dmp';
import { DmpReference } from '@app/core/model/dmp/dmp-reference';
import { ReferenceType } from '@app/core/model/reference-type/reference-type';
import { Reference } from '@app/core/model/reference/reference';
import { RecentActivityItemLookup } from '@app/core/query/recent-activity-item-lookup.lookup';
import { AuthService } from '@app/core/services/auth/auth.service';
import { DashboardService } from '@app/core/services/dashboard/dashboard.service';
import { MatomoService } from '@app/core/services/matomo/matomo-service';
import { EnumUtils } from '@app/core/services/utilities/enum-utils.service';
import { BaseComponent } from '@common/base/base.component';
import { debounceTime, takeUntil } from 'rxjs/operators';
import { nameof } from 'ts-simple-nameof';
import { DescriptionStatus } from '@app/core/common/enum/description-status';
import { AppPermission } from '@app/core/common/enum/permission.enum';
import { DmpBlueprint, DmpBlueprintDefinition, DmpBlueprintDefinitionSection } from '@app/core/model/dmp-blueprint/dmp-blueprint';

@Component({
	selector: 'app-recent-edited-description-activity',
	templateUrl: './recent-edited-description-activity.component.html',
	styleUrls: ['./recent-edited-description-activity.component.scss']
})
export class RecentEditedDescriptionActivityComponent extends BaseComponent implements OnInit {

	lookup: RecentActivityItemLookup = new RecentActivityItemLookup();
	pageSize: number = 5;
	pageLessSize= this.pageSize;
	listingItems: Description[] = [];


	public formGroup = new UntypedFormBuilder().group({
		like: new UntypedFormControl(),
		order: new UntypedFormControl()
	});
	publicMode = false;

	order = RecentActivityOrder;

	totalCount: number;
	startIndex: number = 0;
	offsetLess: number = 0;
	page: number = 1;
	@Input() isActive: boolean = false;

	constructor(
		private route: ActivatedRoute,
		private router: Router,
		public enumUtils: EnumUtils,
		private authentication: AuthService,
		private dashboardService: DashboardService,
		private location: Location,
		private matomoService: MatomoService
	) {
		super();
	}

	ngOnInit() {
		this.matomoService.trackPageView('Recent DMP Activity');
		this.route.queryParams.subscribe(params => {
			if (this.isActive) {
				let page = (params['page'] === undefined) ? 1 : +params['page'];
				this.page = (page <= 0) ? 1 : page;

				this.startIndex = (this.page - 1) * this.pageSize;
				if (this.page > 1) {
					this.offsetLess = (this.page - 2) * this.pageSize;
				}

				let order = params['order'];
				if (this.isAuthenticated()) {
					if (order === undefined || (order != this.order.UpdatedAt && order != this.order.Label && order != this.order.Status)) {
						order = this.order.UpdatedAt;
					}
				} else {
					//TODO refactor
					// if (order === undefined || (order != this.order.PUBLISHED && order != this.order.LABEL)) {
					// 	order = this.order.PUBLISHED;
					// }
				}
				this.formGroup.get('order').setValue(order);

				let keyword = (params['keyword'] === undefined || params['keyword'].length <= 0) ? "" : params['keyword'];
				this.formGroup.get("like").setValue(keyword);

				this.updateUrl();
			}
		});

		this.formGroup.get('like').valueChanges
		.pipe(takeUntil(this._destroyed), debounceTime(500))
		.subscribe(x => {
			console.log('valueChanges');
			this.refresh()
		});
		this.formGroup.get('order').valueChanges
		.pipe(takeUntil(this._destroyed))
		.subscribe(x => this.refresh());

		this.refresh();
	}

	ngOnChanges() {
		if (this.isActive) {
			this.updateUrl();
		}
	}

	updateUrl() {
		let parameters = "?type=descriptions" +
			(this.page != 1 ? "&page=" + this.page : "") +
			//TODO refactor
			//(((this.formGroup.get("order").value != this.order.MODIFIED && !this.publicMode) || (this.formGroup.get("order").value != this.order.PUBLISHED && this.publicMode)) ? "&order=" + this.formGroup.get("order").value : "") +
			(this.formGroup.get("like").value ? ("&keyword=" + this.formGroup.get("like").value) : "");
		this.location.go(this.router.url.split('?')[0] + parameters);
	}

	public isAuthenticated(): boolean {
		return this.authentication.currentAccountIsAuthenticated();
	}

	refresh(): void {
		this.lookup.page = { size: this.pageSize, offset: 0 };
		this.lookup.onlyDescription = true;
		this.lookup.orderField = this.formGroup.get('order').value ??this.order.UpdatedAt;
		this.lookup.like = this.formGroup.get('like').value;
		this.lookup.project = {

			fields: [
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.status)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.updatedAt)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.authorizationFlags), AppPermission.EditDescription].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.authorizationFlags), AppPermission.DeleteDescription].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.authorizationFlags), AppPermission.InviteDmpUsers].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.descriptionTemplate), nameof<DescriptionTemplate>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.descriptionTemplate), nameof<DescriptionTemplate>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.descriptionTemplate), nameof<DescriptionTemplate>(x => x.groupId)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.accessType)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.blueprint), nameof<DmpBlueprint>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.blueprint), nameof<DmpBlueprint>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.blueprint), nameof<DmpBlueprint>(x => x.definition), nameof<DmpBlueprintDefinition>(x => x.sections), nameof<DmpBlueprintDefinitionSection>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.blueprint), nameof<DmpBlueprint>(x => x.definition), nameof<DmpBlueprintDefinition>(x => x.sections), nameof<DmpBlueprintDefinitionSection>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpUsers), nameof<DmpUser>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpUsers), nameof<DmpUser>(x => x.user.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpUsers), nameof<DmpUser>(x => x.role)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.reference), nameof<Reference>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.reference), nameof<Reference>(x => x.label)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.reference), nameof<Reference>(x => x.type), nameof<ReferenceType>(x => x.id)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.reference), nameof<Reference>(x => x.reference)].join('.'),
				[nameof<RecentActivityItem>(x => x.description), nameof<Description>(x => x.dmp), nameof<Dmp>(x => x.dmpReferences), nameof<DmpReference>(x => x.isActive)].join('.'),
			]
		};

		this.getItems();
	}

	loadMore() {
		this.pageSize = this.pageSize + this.pageLessSize;
		this.lookup.page = { size: this.pageSize, offset: 0 };	
		this.getItems();
	}

	loadLess() {
		this.pageSize = this.pageSize - this.pageLessSize;
		this.lookup.page = { size: this.pageSize, offset: 0 };	
		this.getItems();
	}

	getItems(){
		this.listingItems = [];

		this.dashboardService
			.getMyRecentActivityItems(this.lookup)
			.pipe(takeUntil(this._destroyed))
			.subscribe(response => {
				response.forEach(item => {
					if (item.description){
						if (item.description.status != DescriptionStatus.Canceled) this.listingItems.push(item.description);
					}
				})
			});
	}
}
