package gr.cite.notification.web.controllers;

import gr.cite.notification.audit.AuditableAction;
import gr.cite.notification.authorization.AuthorizationFlags;
import gr.cite.notification.data.NotificationTemplateEntity;
import gr.cite.notification.model.NotificationTemplate;
import gr.cite.notification.model.builder.NotificationTemplateBuilder;
import gr.cite.notification.model.censorship.NotificationTemplateCensor;
import gr.cite.notification.model.persist.NotificationTemplatePersist;
import gr.cite.notification.query.NotificationTemplateQuery;
import gr.cite.notification.query.lookup.NotificationTemplateLookup;
import gr.cite.notification.service.notificationtemplate.NotificationTemplateService;
import gr.cite.notification.web.model.QueryResult;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.censor.CensorFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.exception.MyForbiddenException;
import gr.cite.tools.exception.MyNotFoundException;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.logging.MapLogEntry;
import gr.cite.tools.validation.ValidationFilterAnnotation;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.web.bind.annotation.*;

import javax.management.InvalidApplicationException;
import jakarta.transaction.Transactional;
import java.util.*;

@RestController
@RequestMapping(path = "api/notification/notification-template")
public class NotificationTemplateController {
	private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(NotificationTemplateController.class));

	private final BuilderFactory builderFactory;
	private final AuditService auditService;
	private final NotificationTemplateService notificationTemplateService;
	private final CensorFactory censorFactory;
	private final QueryFactory queryFactory;
	private final MessageSource messageSource;

	@Autowired
	public NotificationTemplateController(BuilderFactory builderFactory,
                                          AuditService auditService,
										  NotificationTemplateService notificationTemplateService,
										  CensorFactory censorFactory,
                                          QueryFactory queryFactory,
                                          MessageSource messageSource) {
		this.builderFactory = builderFactory;
		this.auditService = auditService;
		this.notificationTemplateService = notificationTemplateService;
		this.censorFactory = censorFactory;
		this.queryFactory = queryFactory;
		this.messageSource = messageSource;
	}

	@PostMapping("query")
	public QueryResult<NotificationTemplate> query(@RequestBody NotificationTemplateLookup lookup) throws MyApplicationException, MyForbiddenException {
		logger.debug("querying {}", NotificationTemplate.class.getSimpleName());

		this.censorFactory.censor(NotificationTemplateCensor.class).censor(lookup.getProject());

		NotificationTemplateQuery query = lookup.enrich(this.queryFactory).authorize(AuthorizationFlags.OwnerOrPermission);
		List<NotificationTemplateEntity> data = query.collectAs(lookup.getProject());
		List<NotificationTemplate> models = this.builderFactory.builder(NotificationTemplateBuilder.class).authorize(AuthorizationFlags.OwnerOrPermission).build(lookup.getProject(), data);
		long count = (lookup.getMetadata() != null && lookup.getMetadata().getCountAll()) ? query.count() : models.size();

		this.auditService.track(AuditableAction.Notification_Template_Query, "lookup", lookup);

		return new QueryResult<>(models, count);
	}

	@GetMapping("{id}")
	public NotificationTemplate get(@PathVariable UUID id, FieldSet fieldSet, Locale locale) throws MyApplicationException, MyForbiddenException, MyNotFoundException {
		logger.debug(new MapLogEntry("retrieving" + NotificationTemplate.class.getSimpleName()).And("id", id).And("fields", fieldSet));

		this.censorFactory.censor(NotificationTemplateCensor.class).censor(fieldSet);

		NotificationTemplateQuery query = this.queryFactory.query(NotificationTemplateQuery.class).authorize(AuthorizationFlags.OwnerOrPermission).ids(id);
		NotificationTemplate model = this.builderFactory.builder(NotificationTemplateBuilder.class).authorize(AuthorizationFlags.OwnerOrPermission).build(fieldSet, query.first());
		if (model == null)
			throw new MyNotFoundException(messageSource.getMessage("General_ItemNotFound", new Object[]{id, NotificationTemplate.class.getSimpleName()}, LocaleContextHolder.getLocale()));

		this.auditService.track(AuditableAction.Notification_Template_Lookup, Map.ofEntries(
				new AbstractMap.SimpleEntry<String, Object>("id", id),
				new AbstractMap.SimpleEntry<String, Object>("fields", fieldSet)
		));

		return model;
	}

	@PostMapping("persist")
	@Transactional
	@ValidationFilterAnnotation(validator = NotificationTemplatePersist.NotificationTemplatePersistValidator.ValidatorName, argumentName = "model")
	public NotificationTemplate persist(@RequestBody NotificationTemplatePersist model, FieldSet fieldSet) throws MyApplicationException, MyForbiddenException, MyNotFoundException, InvalidApplicationException {
		logger.debug(new MapLogEntry("persisting" + NotificationTemplate.class.getSimpleName()).And("model", model).And("fieldSet", fieldSet));

		NotificationTemplate persisted = this.notificationTemplateService.persist(model, fieldSet);

		this.auditService.track(AuditableAction.Notification_Template_Persist, Map.ofEntries(
				new AbstractMap.SimpleEntry<String, Object>("model", model),
				new AbstractMap.SimpleEntry<String, Object>("fields", fieldSet)
		));

		return persisted;
	}

	@DeleteMapping("{id}")
	@Transactional
	public void delete(@PathVariable("id") UUID id) throws MyForbiddenException, InvalidApplicationException {
		logger.debug(new MapLogEntry("deleting" + NotificationTemplate.class.getSimpleName()).And("id", id));

		this.notificationTemplateService.deleteAndSave(id);

		this.auditService.track(AuditableAction.Notification_Template_Delete, "id", id);
	}

}
