package gr.cite.notification.query;

import gr.cite.commons.web.authz.service.AuthorizationService;
import gr.cite.notification.authorization.AuthorizationFlags;
import gr.cite.notification.authorization.Permission;
import gr.cite.notification.common.enums.ContactInfoType;
import gr.cite.notification.common.enums.IsActive;
import gr.cite.notification.common.scope.user.UserScope;
import gr.cite.notification.data.UserContactInfoEntity;
import gr.cite.notification.data.UserEntity;
import gr.cite.notification.model.User;
import gr.cite.notification.model.UserContactInfo;
import gr.cite.tools.data.query.FieldResolver;
import gr.cite.tools.data.query.QueryBase;
import gr.cite.tools.data.query.QueryContext;
import jakarta.persistence.Tuple;
import jakarta.persistence.criteria.CriteriaBuilder;
import jakarta.persistence.criteria.Predicate;
import org.springframework.stereotype.Component;
import org.springframework.web.context.annotation.RequestScope;

import java.time.Instant;
import java.util.*;

@Component
@RequestScope
public class UserContactInfoQuery extends QueryBase<UserContactInfoEntity> {

    private Collection<UUID> ids;

    private Collection<UUID> userIds;

    private Collection<UUID> excludedUserIds;

    private Collection<IsActive> isActives;

    private Collection<ContactInfoType> type;

    private Collection<String> values;

    private UserQuery userQuery;

    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    private final UserScope userScope;

    private final AuthorizationService authService;

    public UserContactInfoQuery(
            UserScope userScope,
            AuthorizationService authService
    ) {
        this.userScope = userScope;
        this.authService = authService;
    }

    public UserContactInfoQuery ids(UUID value) {
        this.ids = List.of(value);
        return this;
    }

    public UserContactInfoQuery ids(UUID... value) {
        this.ids = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery ids(Collection<UUID> values) {
        this.ids = values;
        return this;
    }

    public UserContactInfoQuery userIds(UUID value) {
        this.userIds = List.of(value);
        return this;
    }

    public UserContactInfoQuery userIds(UUID... value) {
        this.userIds = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery userIds(Collection<UUID> values) {
        this.userIds = values;
        return this;
    }

    public UserContactInfoQuery excludedUserIds(Collection<UUID> values) {
        this.excludedUserIds = values;
        return this;
    }

    public UserContactInfoQuery excludedUserIds(UUID value) {
        this.excludedUserIds = List.of(value);
        return this;
    }

    public UserContactInfoQuery excludedUserIds(UUID... value) {
        this.excludedUserIds = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery isActive(IsActive value) {
        this.isActives = List.of(value);
        return this;
    }

    public UserContactInfoQuery isActive(IsActive... value) {
        this.isActives = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery isActive(Collection<IsActive> values) {
        this.isActives = values;
        return this;
    }

    public UserContactInfoQuery type(ContactInfoType value) {
        this.type = List.of(value);
        return this;
    }

    public UserContactInfoQuery type(ContactInfoType... value) {
        this.type = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery type(Collection<ContactInfoType> values) {
        this.type = values;
        return this;
    }

    public UserContactInfoQuery values(String value) {
        this.values = List.of(value);
        return this;
    }

    public UserContactInfoQuery values(String... value) {
        this.values = Arrays.asList(value);
        return this;
    }

    public UserContactInfoQuery values(Collection<String> values) {
        this.values = values;
        return this;
    }

    public UserContactInfoQuery userSubQuery(UserQuery subQuery) {
        this.userQuery = subQuery;
        return this;
    }

    public UserContactInfoQuery authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    protected Boolean isFalseQuery() {
        return this.isEmpty(this.userIds) || this.isEmpty(this.excludedUserIds)  || this.isEmpty(this.isActives)
                || this.isEmpty(this.type) || this.isEmpty(this.excludedUserIds) || this.isFalseQuery(this.userQuery);
    }

    @Override
    protected Class<UserContactInfoEntity> entityClass() {
        return UserContactInfoEntity.class;
    }

    @Override
    protected <X, Y> Predicate applyAuthZ(QueryContext<X, Y> queryContext) {
        if (this.authorize.contains(AuthorizationFlags.None))
            return null;
        if (this.authorize.contains(AuthorizationFlags.Permission) && this.authService.authorize(Permission.BrowseUserContactInfo))
            return null;
        UUID ownerId = null;
        if (this.authorize.contains(AuthorizationFlags.Owner))
            ownerId = this.userScope.getUserIdSafe();

        List<Predicate> predicates = new ArrayList<>();
        if (ownerId != null) {
            predicates.add(queryContext.CriteriaBuilder.equal(queryContext.Root.get(UserContactInfoEntity._userId), ownerId));
        }
        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return queryContext.CriteriaBuilder.or(); //Creates a false query
        }
    }

    @Override
    protected <X, Y> Predicate applyFilters(QueryContext<X, Y> queryContext) {
        List<Predicate> predicates = new ArrayList<>();
        if (this.isActives != null) {
            predicates.add(queryContext.Root.get(UserContactInfoEntity._isActive).in(isActives));
        }
        if (this.userIds != null) {
            CriteriaBuilder.In<UUID> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(UserContactInfoEntity._userId));
            for (UUID item : this.userIds)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.excludedUserIds != null) {
            CriteriaBuilder.In<UUID> notInClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(UserContactInfoEntity._userId));
            for (UUID item : this.excludedUserIds)
                notInClause.value(item);
            predicates.add(notInClause.not());
        }
        if (this.type != null) {
            CriteriaBuilder.In<ContactInfoType> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(UserContactInfoEntity._type));
            for (ContactInfoType item : this.type)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.values != null) {
            CriteriaBuilder.In<String> inClause = queryContext.CriteriaBuilder.in(queryContext.Root.get(UserContactInfoEntity._value));
            for (String item : this.values)
                inClause.value(item);
            predicates.add(inClause);
        }
        if (this.userQuery != null) {
            QueryContext<UserEntity, UUID> subQuery = this.applySubQuery(this.userQuery, queryContext, UUID.class, root -> root.get(UserContactInfoEntity._userId));
            predicates.add(queryContext.CriteriaBuilder.in(queryContext.Root.get(UserContactInfoEntity._userId)).value(subQuery.Query));
        }

        if (!predicates.isEmpty()) {
            Predicate[] predicatesArray = predicates.toArray(new Predicate[0]);
            return queryContext.CriteriaBuilder.and(predicatesArray);
        } else {
            return null;
        }
    }

    @Override
    protected String fieldNameOf(FieldResolver item) {
        if (item.match(UserContactInfo._id))
            return UserContactInfoEntity._id;
        else if (item.match(UserContactInfo._type))
            return UserContactInfoEntity._type;
        else if (item.match(UserContactInfo._value))
            return UserContactInfoEntity._value;
        else if (item.match(UserContactInfo._ordinal))
            return UserContactInfoEntity._ordinal;
        else if (item.match(UserContactInfo._createdAt))
            return UserContactInfoEntity._createdAt;
        else if (item.match(UserContactInfo._updatedAt))
            return UserContactInfoEntity._updatedAt;
        else if (item.match(UserContactInfo._isActive))
            return UserContactInfoEntity._isActive;
        else if (item.match(UserContactInfo._user, User._id))
            return UserContactInfoEntity._userId;
        else if (item.prefix(UserContactInfo._user))
            return UserContactInfoEntity._userId;
        else
            return null;
    }

    @Override
    protected UserContactInfoEntity convert(Tuple tuple, Set<String> columns) {
        UserContactInfoEntity item = new UserContactInfoEntity();
        item.setId(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._id, UUID.class));
        item.setType(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._type, ContactInfoType.class));
        item.setValue(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._value, String.class));
        item.setOrdinal(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._ordinal, Integer.class));
        item.setUserId(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._userId, UUID.class));
        item.setCreatedAt(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._createdAt, Instant.class));
        item.setUpdatedAt(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._updatedAt, Instant.class));
        item.setIsActive(QueryBase.convertSafe(tuple, columns, UserContactInfoEntity._isActive, IsActive.class));
        return item;
    }

}
